@@ -93,6 +93,20 @@ extern int ptymode;
 extern char *ptyterm;
 extern int ptymenukey;
 #endif
+#ifndef	_NOLOGGING
+extern char *logfile;
+extern int logsize;
+# ifndef	NOSYSLOG
+extern int usesyslog;
+# endif
+extern int loglevel;
+# ifndef	NOUID
+extern int rootloglevel;
+# endif
+#endif	/* !_NOLOGGING */
+#if	FD >= 2
+extern int thruargs;
+#endif
 extern int wheader;
 extern char fullpath[];
 extern char *origpath;
@@ -165,6 +179,8 @@ typedef struct _envtable {
 	u_char type;
 } envtable;
 
+#define	env_str(n)	(&(envlist[n].env[FDESIZ]))
+#define	fdenv_str(n)	(envlist[n].env)
 #ifdef	FORCEDSTDC
 #define	def_str(n)	(envlist[n].def.str)
 #define	def_num(n)	(envlist[n].def.num)
@@ -196,8 +212,9 @@ typedef struct _envtable {
 #define	T_KIN		14
 #define	T_KOUT		15
 #define	T_KNAM		16
-#define	T_OCTAL		17
-#define	T_KEYCODE	18
+#define	T_KTERM		17
+#define	T_OCTAL		18
+#define	T_KEYCODE	19
 
 #ifndef	_NOKANJIFCONV
 typedef struct _pathtable {
@@ -393,6 +410,21 @@ static CONST envtable envlist[] = {
 	{"FD_PTYTERM", &ptyterm, DEFVAL(PTYTERM), PTYTM_E, T_CHARP},
 	{"FD_PTYMENUKEY", &ptymenukey, DEFVAL(PTYMENUKEY), PTYKY_E, T_KEYCODE},
 #endif
+#ifndef	_NOLOGGING
+	{"FD_LOGFILE", &logfile, DEFVAL(LOGFILE), LGFIL_E, T_PATH},
+	{"FD_LOGSIZE", &logsize, DEFVAL(LOGSIZE), LGSIZ_E, T_INT},
+# ifndef	NOSYSLOG
+	{"FD_USESYSLOG", &usesyslog, DEFVAL(USESYSLOG), USYLG_E, T_BOOL},
+# endif
+	{"FD_LOGLEVEL", &loglevel, DEFVAL(LOGLEVEL), LGLVL_E, T_INT},
+# ifndef	NOUID
+	{"FD_ROOTLOGLEVEL", &rootloglevel,
+		DEFVAL(ROOTLOGLEVEL), RLGLV_E, T_INT},
+# endif
+#endif	/* !_NOLOGGING */
+#if	FD >= 2
+	{"FD_THRUARGS", &thruargs, DEFVAL(THRUARGS), THARG_E, T_BOOL},
+#endif
 #if	!defined (_NOKANJICONV) || (defined (FD) && !defined (_NODOSDRIVE))
 	{"FD_UNICODEBUFFER", &unicodebuffer,
 		DEFVAL(UNICODEBUFFER), UNBF_E, T_BOOL},
@@ -407,6 +439,10 @@ static CONST envtable envlist[] = {
 #ifndef	_NOKANJICONV
 	{"FD_INPUTKCODE", &inputkcode, DEFVAL(NOCNV), IPKC_E, T_KIN},
 #endif
+#if	!defined (_NOKANJICONV) && !defined (_NOPTY)
+	{"FD_PTYINKCODE", &ptyinkcode, DEFVAL(NOCNV), PIKC_E, T_KTERM},
+	{"FD_PTYOUTKCODE", &ptyoutkcode, DEFVAL(NOCNV), POKC_E, T_KTERM},
+#endif
 #ifndef	_NOKANJIFCONV
 	{"FD_FNAMEKCODE", &fnamekcode, DEFVAL(NOCNV), FNKC_E, T_KNAM},
 	{"FD_SJISPATH", &sjispath, DEFVAL(SJISPATH), SJSP_E, T_PATHS},
@@ -425,7 +461,7 @@ static CONST envtable envlist[] = {
 	{"FD_NOCONVPATH", &noconvpath, DEFVAL(NOCONVPATH), NCVP_E, T_PATHS},
 #endif	/* !_NOKANJIFCONV */
 };
-#define	ENVLISTSIZ	((int)(sizeof(envlist) / sizeof(envtable)))
+#define	ENVLISTSIZ	arraysize(envlist)
 
 #ifndef	_NOKANJIFCONV
 static pathtable pathlist[] = {
@@ -434,7 +470,7 @@ static pathtable pathlist[] = {
 	{&progpath, NULL, NOCNV, P_STABLE},
 	{&unitblpath, NULL, NOCNV, P_STABLE},
 };
-#define	PATHLISTSIZ	((int)(sizeof(pathlist) / sizeof(pathtable)))
+#define	PATHLISTSIZ	arraysize(pathlist)
 # ifndef	_NOSPLITWIN
 #  ifndef	_NOARCHIVE
 static pathtable archlist[MAXWINDOWS];
@@ -455,7 +491,7 @@ static CONST devinfo mediadescr[] = {
 	{0xf8, "HDD(PC98)", 'N', 98, 0},
 #  endif
 };
-#define	MEDIADESCRSIZ	((int)(sizeof(mediadescr) / sizeof(devinfo)))
+#define	MEDIADESCRSIZ	arraysize(mediadescr)
 # endif	/* _USEDOSEMU */
 static int cs_item = 0;
 static int cs_max = 0;
@@ -488,7 +524,7 @@ char *s;
 {
 	int n;
 
-	if (!sscanf2(s, "%o%$", &n)) return(-1);
+	if (!sscanf2(s, "%<o%$", &n)) return(-1);
 	n &= 0777;
 
 	return(n);
@@ -501,7 +537,7 @@ int no;
 	char *cp;
 	int n;
 
-	cp = getenv2(envlist[no].env);
+	cp = getenv2(fdenv_str(no));
 	switch (envlist[no].type) {
 		case T_BOOL:
 			if (!cp) n = def_num(no);
@@ -599,6 +635,7 @@ int no;
 		case T_KIN:
 		case T_KOUT:
 		case T_KNAM:
+		case T_KTERM:
 			n = (1 << (envlist[no].type - T_KIN));
 			*((int *)(envlist[no].var)) = getlang(cp, n);
 			break;
@@ -635,7 +672,10 @@ int stable;
 
 	if (!stable) /*EMPTY*/;
 	else if (pp -> lang == NOCNV) {
-		pp -> last = NULL;
+		if (!(pp -> flags & P_STABLE)) {
+			if (pp -> last) free(pp -> last);
+			pp -> last = NULL;
+		}
 		stable = 0;
 	}
 	else if (pp -> flags & P_STABLE) /*EMPTY*/;
@@ -668,7 +708,10 @@ pathtable *pp;
 		}
 		if (kanjierrno) pp -> lang = DEFCODE;
 	}
-	pp -> last = strdup2(path);
+	if (!(pp -> flags & P_STABLE)) {
+		if (pp -> last) free(pp -> last);
+		pp -> last = strdup2(path);
+	}
 }
 
 static VOID NEAR savepathlang(VOID_A)
@@ -705,6 +748,8 @@ static VOID NEAR evalpathlang(VOID_A)
 # endif	/* !_NOSPLITWIN */
 # ifndef	_NOPTY
 	changekcode();
+	changeinkcode();
+	changeoutkcode();
 # endif
 }
 #endif	/* !_NOKANJIFCONV */
@@ -748,6 +793,9 @@ VOID evalenv(VOID_A)
 #endif
 	evalheader();
 	errno = duperrno;
+#if	!MSDOS && !defined (_NOORIGSHELL)
+	if (autosavetty) savestdio(0);
+#endif
 }
 
 #ifdef	DEBUG
@@ -755,6 +803,12 @@ VOID freeenvpath(VOID_A)
 {
 	int i;
 
+# ifndef	_NOKANJIFCONV
+	for (i = 0; i < PATHLISTSIZ; i++) {
+		if (pathlist[i].last) free(pathlist[i].last);
+		pathlist[i].last = NULL;
+	}
+# endif
 	for (i = 0; i < ENVLISTSIZ; i++) switch (envlist[i].type) {
 		case T_PATH:
 		case T_PATHS:
@@ -800,7 +854,7 @@ int y, w;
 {
 	Xlocate(0, y);
 	Xputterm(T_STANDOUT);
-	Xcprintf2("%*s", w, "");
+	cputspace(w);
 	Xputterm(END_STANDOUT);
 }
 
@@ -1057,9 +1111,8 @@ char **list;
 		savepathlang();
 #endif
 		for (i = 0; i < ENVLISTSIZ; i++) {
-			setenv2(envlist[i].env, list[i * 2], 0);
-			setenv2(&(envlist[i].env[FDESIZ]),
-				list[i * 2 + 1], 0);
+			setenv2(fdenv_str(i), list[i * 2], 0);
+			setenv2(env_str(i), list[i * 2 + 1], 0);
 			_evalenv(i);
 		}
 #ifndef	_NOKANJIFCONV
@@ -1070,10 +1123,8 @@ char **list;
 	else {
 		list = (char **)malloc2(ENVLISTSIZ * 2 * sizeof(char *));
 		for (i = 0; i < ENVLISTSIZ; i++) {
-			list[i * 2] = strdup2(getshellvar(envlist[i].env, -1));
-			list[i * 2 + 1] =
-				strdup2(getshellvar(&(envlist[i].env[FDESIZ]),
-					-1));
+			list[i * 2] = strdup2(getshellvar(fdenv_str(i), -1));
+			list[i * 2 + 1] = strdup2(getshellvar(env_str(i), -1));
 		}
 	}
 
@@ -1089,9 +1140,9 @@ static VOID NEAR cleanupenv(VOID_A)
 	savepathlang();
 #endif
 	for (i = 0; i < ENVLISTSIZ; i++) {
-		setenv2(envlist[i].env, NULL, 0);
+		setenv2(fdenv_str(i), NULL, 0);
 		cp = (envlist[i].type == T_CHARP) ? def_str(i) : NULL;
-		setenv2(&(envlist[i].env[FDESIZ]), cp, 0);
+		setenv2(env_str(i), cp, 0);
 		_evalenv(i);
 	}
 #ifndef	_NOKANJIFCONV
@@ -1249,6 +1300,7 @@ int no;
 		case T_KIN:
 		case T_KOUT:
 		case T_KNAM:
+		case T_KTERM:
 			str[NOCNV] = VNCNV_K;
 			str[ENG] = VENG_K;
 #  ifndef	_NOKANJICONV
@@ -1275,18 +1327,18 @@ int no;
 # endif
 # ifndef	_NOPTY
 		case T_KEYCODE:
-			cp = getenv2(envlist[no].env);
+			cp = getenv2(fdenv_str(no));
 			break;
 # endif
 		default:
-			if (!(cp = getenv2(envlist[no].env))) cp = def_str(no);
+			if (!(cp = getenv2(fdenv_str(no)))) cp = def_str(no);
 			break;
 	}
 
-	if (!getenv2(envlist[no].env))
+	if (!getenv2(fdenv_str(no)))
 		cp = new = strcatalloc(strdup2(cp),
 			(cp && *cp) ? VDEF_K : VUDEF_K);
-	Xcprintf2("%-*.*k", MAXCUSTVAL, MAXCUSTVAL, cp);
+	cputstr(MAXCUSTVAL, cp);
 	n = strlen2(cp);
 	if (n > MAXCUSTVAL - 1) n = MAXCUSTVAL - 1;
 	if (new) free(new);
@@ -1303,7 +1355,7 @@ int no;
 	for (n = 0; n < MAXSELECTSTRS; n++) val[n] = n;
 
 	new = NULL;
-	env = &(envlist[no].env[FDESIZ]);
+	env = env_str(no);
 	switch (envlist[no].type) {
 		case T_BOOL:
 			n = (*((int *)(envlist[no].var))) ? 1 : 0;
@@ -1344,7 +1396,7 @@ int no;
 			break;
 		case T_PATH:
 		case T_PATHS:
-			if (!(cp = getenv2(envlist[no].env))) cp = def_str(no);
+			if (!(cp = getenv2(fdenv_str(no)))) cp = def_str(no);
 			new = inputcustenvstr(env, 1, cp, HST_PATH);
 			if (new == (char *)-1) return(0);
 			cp = new;
@@ -1508,6 +1560,7 @@ int no;
 		case T_KIN:
 		case T_KNAM:
 		case T_KOUT:
+		case T_KTERM:
 			tmp = 0;
 			str[tmp] = VNCNV_K;
 			val[tmp++] = NOCNV;
@@ -1594,7 +1647,7 @@ int no;
 				n += p;
 			}
 #  endif	/* !_NOKANJICONV */
-			str[NOCNV] = "";
+			str[NOCNV] = nullstr;
 #  if	!defined (_NOENGMES) && !defined (_NOJPNMES)
 			str[ENG] = "C";
 #  endif
@@ -1632,7 +1685,7 @@ int no;
 # endif	/* FD >= 2 */
 # ifndef	_NOPTY
 		case T_KEYCODE:
-			if (!(cp = getenv2(envlist[no].env))) cp = "";
+			if (!(cp = getenv2(fdenv_str(no)))) cp = nullstr;
 			new = inputcustenvstr(env, 0, cp, -1);
 			if (new == (char *)-1) return(0);
 			if (new && getkeycode(new, 0) < 0) {
@@ -1644,16 +1697,17 @@ int no;
 			break;
 # endif	/* !_NOPTY */
 		default:
-			if (!(cp = getenv2(envlist[no].env))) cp = def_str(no);
+			if (!(cp = getenv2(fdenv_str(no)))) cp = def_str(no);
 			new = inputcustenvstr(env, 0, cp, -1);
 			if (new == (char *)-1) return(0);
 			cp = new;
 			break;
 	}
 
-	if (getshellvar(envlist[no].env, -1)) setenv2(envlist[no].env, cp, 0);
-	else setenv2(env, cp, 0);
+	if (getshellvar(fdenv_str(no), -1)) n = setenv2(fdenv_str(no), cp, 0);
+	else n = setenv2(env, cp, 0);
 	if (new) free(new);
+	if (n < 0) warning(-1, env);
 #ifndef	_NOKANJIFCONV
 	savepathlang();
 #endif
@@ -1675,11 +1729,11 @@ FILE *fp;
 
 	for (i = n = 0; i < ENVLISTSIZ; i++) {
 		if ((!flaglist || !(flaglist[i] & 2))
-		&& (cp = getshellvar(&(envlist[i].env[FDESIZ]), -1))
+		&& (cp = getshellvar(env_str(i), -1))
 		&& (envlist[i].type != T_CHARP || strenvcmp(cp, def_str(i))))
 			n++;
 		if ((!flaglist || !(flaglist[i] & 1))
-		&& getshellvar(envlist[i].env, -1))
+		&& getshellvar(fdenv_str(i), -1))
 			n++;
 	}
 	if (!n || !fp) return(n);
@@ -1688,16 +1742,16 @@ FILE *fp;
 	fputs("# shell variables definition\n", fp);
 	for (i = 0; i < ENVLISTSIZ; i++) {
 		if ((!flaglist || !(flaglist[i] & 2))
-		&& (cp = getshellvar(&(envlist[i].env[FDESIZ]), -1))
+		&& (cp = getshellvar(env_str(i), -1))
 		&& (envlist[i].type != T_CHARP || strenvcmp(cp, def_str(i)))) {
 			cp = killmeta(cp);
-			fprintf2(fp, "%s=%s\n", &(envlist[i].env[FDESIZ]), cp);
+			fprintf2(fp, "%s=%s\n", env_str(i), cp);
 			free(cp);
 		}
 		if ((!flaglist || !(flaglist[i] & 1))
-		&& (cp = getshellvar(envlist[i].env, -1))) {
+		&& (cp = getshellvar(fdenv_str(i), -1))) {
 			cp = killmeta(cp);
-			fprintf2(fp, "%s=%s\n", envlist[i].env, cp);
+			fprintf2(fp, "%s=%s\n", fdenv_str(i), cp);
 			free(cp);
 		}
 	}
@@ -1716,10 +1770,8 @@ FILE *fp;
 
 	for (n = 0; argv[n]; n++) {
 		for (i = 0; i < ENVLISTSIZ; i++) {
-			ident = envlist[i].env;
-			if (!strnenvcmp(argv[n], ident, len[n])) break;
-			ident += FDESIZ;
-			if (!strnenvcmp(argv[n], ident, len[n])) break;
+			if (!strnenvcmp(argv[n], fdenv_str(i), len[n])) break;
+			if (!strnenvcmp(argv[n], env_str(i), len[n])) break;
 		}
 		if (i < ENVLISTSIZ) break;
 	}
@@ -1732,12 +1784,12 @@ FILE *fp;
 		f = 0;	/* fake for -Wuninitialized */
 #  endif
 		for (i = 0; i < ENVLISTSIZ; i++) {
-			ident = envlist[i].env;
+			ident = fdenv_str(i);
 			if (!strnenvcmp(argv[n], ident, len[n])) {
 				f = 1;
 				break;
 			}
-			ident += FDESIZ;
+			ident = env_str(i);
 			if (!strnenvcmp(argv[n], ident, len[n])) {
 				f = 2;
 				break;
@@ -1797,10 +1849,8 @@ FILE *fp;
 
 	for (n = 1; argv[n]; n++) {
 		for (i = 0; i < ENVLISTSIZ; i++) {
-			ident = envlist[i].env;
-			if (!strenvcmp(argv[n], ident)) break;
-			ident += FDESIZ;
-			if (!strenvcmp(argv[n], ident)) break;
+			if (!strenvcmp(argv[n], fdenv_str(i))) break;
+			if (!strenvcmp(argv[n], env_str(i))) break;
 		}
 		if (i < ENVLISTSIZ) break;
 	}
@@ -1813,12 +1863,12 @@ FILE *fp;
 		f = 0;	/* fake for -Wuninitialized */
 #  endif
 		for (i = 0; i < ENVLISTSIZ; i++) {
-			ident = envlist[i].env;
+			ident = fdenv_str(i);
 			if (!strenvcmp(argv[n], ident)) {
 				f = 1;
 				break;
 			}
-			ident += FDESIZ;
+			ident = env_str(i);
 			if (!strenvcmp(argv[n], ident)) {
 				f = 2;
 				break;
@@ -1900,7 +1950,7 @@ int no;
 	int len, width;
 
 	if (bindlist[no].key < 0) {
-		Xcprintf2("%*s", MAXCUSTVAL, "");
+		cputspace(MAXCUSTVAL);
 		return(0);
 	}
 	if (bindlist[no].f_func < FUNCLISTSIZ)
@@ -1913,14 +1963,13 @@ int no;
 	else cp2 = NULL;
 	if (!cp2) {
 		width = MAXCUSTVAL;
-		Xcprintf2("%-*.*k", width, width, cp1);
+		cputstr(width, cp1);
 	}
 	else {
 		width = (MAXCUSTVAL - 1) / 2;
-		Xcprintf2("%-*.*k", width, width, cp1);
+		cputstr(width, cp1);
 		putsep();
-		Xcprintf2("%-*.*k",
-			MAXCUSTVAL - 1 - width, MAXCUSTVAL - 1 - width, cp2);
+		cputstr(MAXCUSTVAL - 1 - width, cp2);
 	}
 	len = strlen2(cp1);
 	if (len > --width) len = width;
@@ -2220,11 +2269,11 @@ int no;
 
 	key.code = keyseqlist[no];
 	if (getkeyseq(&key) < 0 || !(key.len)) {
-		Xcprintf2("%*s", MAXCUSTVAL, "");
+		cputspace(MAXCUSTVAL);
 		return(0);
 	}
 	cp = encodestr(key.str, key.len);
-	Xcprintf2("%-*.*k", MAXCUSTVAL, MAXCUSTVAL, cp);
+	cputstr(MAXCUSTVAL, cp);
 	len = strlen2(cp);
 	if (len > MAXCUSTVAL - 1) len = MAXCUSTVAL - 1;
 	free(cp);
@@ -2475,19 +2524,19 @@ int no;
 	int len, width;
 
 	if (no >= maxlaunch) {
-		Xcprintf2("%*s", MAXCUSTVAL, "");
+		cputspace(MAXCUSTVAL);
 		return(0);
 	}
 	if (!launchlist[no].format) {
 		width = MAXCUSTVAL;
-		Xcprintf2("%-*.*k", width, width, launchlist[no].comm);
+		cputstr(width, launchlist[no].comm);
 	}
 	else {
 		width = (MAXCUSTVAL - 1 - 6) / 2;
-		Xcprintf2("%-*.*k", width, width, launchlist[no].comm);
+		cputstr(width, launchlist[no].comm);
 		putsep();
 		width = MAXCUSTVAL - 1 - 6 - width;
-		Xcprintf2("%-*.*k", width, width, launchlist[no].format[0]);
+		cputstr(width, launchlist[no].format[0]);
 		putsep();
 		Xcprintf2("%-2d", (int)(launchlist[no].topskip));
 		putsep();
@@ -2519,7 +2568,7 @@ launchtable *list;
 	}
 	Xcprintf2("%.*k", len, list -> ext + 1);
 	putsep();
-	Xcprintf2("%-*.*k", MAXCUSTVAL, MAXCUSTVAL, list -> comm);
+	cputstr(MAXCUSTVAL, list -> comm);
 	putsep();
 	fillline(yy + y++, n_column);
 	nf = ni = ne = 0;
@@ -2549,8 +2598,7 @@ launchtable *list;
 				Xputterm(END_STANDOUT);
 				putsep();
 			}
-			Xcprintf2("%-*.*k",
-				MAXCUSTVAL, MAXCUSTVAL, list -> format[nf]);
+			cputstr(MAXCUSTVAL, list -> format[nf]);
 			putsep();
 			nf++;
 		}
@@ -2568,8 +2616,7 @@ launchtable *list;
 				Xputterm(END_STANDOUT);
 				putsep();
 			}
-			Xcprintf2("%-*.*k",
-				MAXCUSTVAL, MAXCUSTVAL, list -> lignore[ni]);
+			cputstr(MAXCUSTVAL, list -> lignore[ni]);
 			putsep();
 			ni++;
 		}
@@ -2587,8 +2634,7 @@ launchtable *list;
 				Xputterm(END_STANDOUT);
 				putsep();
 			}
-			Xcprintf2("%-*.*k",
-				MAXCUSTVAL, MAXCUSTVAL, list -> lerror[ne]);
+			cputstr(MAXCUSTVAL, list -> lerror[ne]);
 			putsep();
 			ne++;
 		}
@@ -2605,7 +2651,7 @@ char *prompt, **var;
 	int i, n, max, val[4];
 
 	max = countvar(var);
-	for (i = 0; i < (int)sizeof(val) / sizeof(int); i++) val[i] = i;
+	for (i = 0; i < arraysize(val); i++) val[i] = i;
 
 	list = (namelist *)malloc2((max + 1) * sizeof(namelist));
 	mes = (char **)malloc2((max + 1) * sizeof(char *));
@@ -2646,8 +2692,7 @@ char *prompt, **var;
 		str[2] = (n < max - 1) ? ARDWN_K : NULL;
 		str[3] = ARDEL_K;
 		i = 0;
-		if (noselect(&i, sizeof(val) / sizeof(int), 0, str, val))
-			continue;
+		if (noselect(&i, arraysize(val), 0, str, val)) continue;
 		if (i == 0) {
 			if (!(tmp = inputcuststr(prompt, 1, cp, -1)))
 				continue;
@@ -2702,17 +2747,17 @@ launchtable *list;
 	str[3] = "-i";
 	str[4] = "-e";
 	str[5] = ARDON_K;
-	for (i = 0; i < (int)sizeof(val) / sizeof(int); i++) val[i] = i;
+	for (i = 0; i < arraysize(val); i++) val[i] = i;
 
 	n = 0;
 	for (;;) {
 		verboselaunch(list);
 		envcaption(ARSEL_K);
-		if (noselect(&n, sizeof(val) / sizeof(int), 0, str, val)) {
+		if (noselect(&n, arraysize(val), 0, str, val)) {
 			if (yesno(ARCAN_K)) return(0);
 			continue;
 		}
-		if (n == (int)sizeof(val) / sizeof(int) - 1) {
+		if (n == arraysize(val) - 1) {
 			if (!(list -> format)) {
 				if (!yesno(ARNOT_K)) continue;
 				freevar(list -> lignore);
@@ -3025,18 +3070,13 @@ int no;
 	int len, width;
 
 	if (no >= maxarchive) {
-		Xcprintf2("%*s", MAXCUSTVAL, "");
+		cputspace(MAXCUSTVAL);
 		return(0);
 	}
 	width = (MAXCUSTVAL - 1) / 2;
-	if (!archivelist[no].p_comm) Xcprintf2("%*s", width, "");
-	else Xcprintf2("%-*.*k", width, width, archivelist[no].p_comm);
+	cputstr(width, archivelist[no].p_comm);
 	putsep();
-	if (!archivelist[no].u_comm)
-		Xcprintf2("%*s", MAXCUSTVAL - 1 - width, "");
-	else Xcprintf2("%-*.*k",
-		MAXCUSTVAL - 1 - width, MAXCUSTVAL - 1 - width,
-		archivelist[no].u_comm);
+	cputstr(MAXCUSTVAL - 1 - width, archivelist[no].u_comm);
 	len = (archivelist[no].p_comm) ? strlen2(archivelist[no].p_comm) : 0;
 	if (len > --width) len = width;
 
@@ -3281,24 +3321,24 @@ int no;
 	int i, len, w1, w2, width;
 
 	if (!fdtype[no].name) {
-		Xcprintf2("%*s", MAXCUSTVAL, "");
+		cputspace(MAXCUSTVAL);
 		return(0);
 	}
 	width = (MAXCUSTVAL - 1) / 2;
-	Xcprintf2("%-*.*k", width, width, fdtype[no].name);
+	cputstr(width, fdtype[no].name);
 	putsep();
 
 	w1 = MAXCUSTVAL - 1 - width;
 #  ifdef	HDDMOUNT
 	if (!fdtype[no].cyl) {
-		char buf[sizeof("HDD98 #offset=") + MAXCOLSCOMMA(3)];
+		char buf[strsize("HDD98 #offset=") + MAXCOLSCOMMA(3) + 1];
 
 		strcpy(buf, "HDD");
 		if (isupper2(fdtype[no].head)) strcat(buf, "98");
 		i = strlen(buf);
-		snprintf2(&(buf[i]), sizeof(buf) - i, " #offset=%'Ld",
+		snprintf2(&(buf[i]), (int)sizeof(buf) - i, " #offset=%'Ld",
 			fdtype[no].offset / fdtype[no].sect);
-		Xcprintf2("%-*.*k", w1, w1, buf);
+		cputstr(w1, buf);
 	}
 	else
 #  endif	/* HDDMOUNT */
@@ -3325,8 +3365,7 @@ int no;
 				break;
 		}
 		w1 -= (w2 + 1) * 3;
-		if (i >= MEDIADESCRSIZ) Xcprintf2("%*s", w1, "");
-		else Xcprintf2("%-*.*k", w1, w1, mediadescr[i].name);
+		cputstr(w1, (i < MEDIADESCRSIZ) ? mediadescr[i].name : NULL);
 	}
 	len = strlen2(fdtype[no].name);
 	if (len > --width) len = width;
@@ -3621,7 +3660,7 @@ int no;
 # else
 	mes[4] = COVWR_K;
 # endif
-	Xcprintf2("%-*.*k", MAXCUSTVAL, MAXCUSTVAL, mes[no]);
+	cputstr(MAXCUSTVAL, mes[no]);
 	len = strlen2(mes[no]);
 	if (len > MAXCUSTVAL - 1) len = MAXCUSTVAL - 1;
 
@@ -3646,7 +3685,7 @@ char *file;
 
 	strcpy(path, file);
 	cp = getbasename(path);
-	len = (int)sizeof(path) - 1 - (cp - path);
+	len = strsize(path) - (cp - path);
 	if (len > MAXTNAMLEN) len = MAXTNAMLEN;
 	genrandname(NULL, 0);
 
@@ -4019,7 +4058,7 @@ int no, y, isstandout;
 	cp = NEWET_K;
 	switch (custno) {
 		case 0:
-			cp = &(envlist[no].env[FDESIZ]);
+			cp = env_str(no);
 			break;
 		case 1:
 			if (bindlist[no].key >= 0)
@@ -4075,7 +4114,7 @@ int no, y, isstandout;
 			cp = name[no];
 			break;
 		default:
-			cp = "";
+			cp = nullstr;
 			break;
 	}
 
@@ -4138,8 +4177,7 @@ static VOID NEAR dispcust(VOID_A)
 					Xcprintf2("%.*k", MAXCUSTVAL, NIMPL_K);
 				if (cs_len[i - start] > MAXCUSTVAL - 1)
 					cs_len[i - start] = MAXCUSTVAL - 1;
-				else Xcprintf2("%*s",
-					MAXCUSTVAL - cs_len[i - start], "");
+				else cputspace(MAXCUSTVAL - cs_len[i - start]);
 				break;
 		}
 		putsep();
