@@ -5,6 +5,7 @@
  */
 
 #include "fd.h"
+#include "funcno.h"
 
 #ifndef	_NOPTY
 
@@ -29,13 +30,16 @@ extern int deletealias __P_((char *));
 #define	sigalrm()	0
 #endif
 
+extern functable funclist[];
 extern int internal_status;
 #ifdef	SIGALRM
 extern int noalrm;
 #endif
+extern int fdmode;
 extern int fdflags;
 extern int wheader;
 extern int ptymode;
+extern int ptyinternal;
 extern int ptymenukey;
 extern p_id_t emupid;
 extern int emufd;
@@ -75,23 +79,25 @@ int options, *statusp;
 static int waitpty(VOID_A)
 {
 	char result[MAXWINDOWS];
-	int i, n, fds[MAXWINDOWS];
+	int i, oldwin, fds[MAXWINDOWS];
 
 	if (lastfunc && (*lastfunc)() < 0) return(-1);
 
-	for (i = 0; i < MAXWINDOWS; i++) fds[i] = ptylist[i].pipe;
-	if (selectpty(-1, fds, result, 0) > 0) {
+	oldwin = win;
+	for (i = 0; i < MAXWINDOWS; i++)
+		fds[i] = (ptylist[i].pid) ? ptylist[i].pipe : -1;
+	if (selectpty(MAXWINDOWS, fds, result, 0) > 0) {
 		for (i = 0; i < MAXWINDOWS; i++)
 			if (result[i] && ptylist[i].pid) recvchild(i);
 	}
 
+	if (win != oldwin) {
+		recvchild(oldwin);
+		return(-2);
+	}
 	if (!(ptylist[win].pid)) return(0);
 	if (ptylist[win].status >= 0) return(-1);
-	if (waitstatus(ptylist[win].pid, WNOHANG, &n) < 0) return(0);
-	ptylist[win].pid = (p_id_t)0;
-	ptylist[win].status = n;
-
-	return(-1);
+	return(checkpty(win));
 }
 
 VOID Xttyiomode(isnl)
@@ -135,6 +141,9 @@ int init;
 	int mode, dupdumbterm;
 
 	dupdumbterm = dumbterm;
+#ifndef	_NOORIGSHELL
+	if (isshptymode()) dumbterm = 1;
+#endif
 	if (isptymode()) dumbterm = 1;
 	mode = termmode(init);
 	dumbterm = dupdumbterm;
@@ -361,6 +370,26 @@ VOID changekcode(VOID_A)
 	sendword(emufd, inputkcode);
 	sendword(emufd, outputkcode);
 }
+
+VOID changeinkcode(VOID_A)
+{
+	ptylist[win].incode = ptyinkcode;
+	if (!emupid) return;
+
+	sendword(emufd, TE_CHANGEINKCODE);
+	sendword(emufd, win);
+	sendword(emufd, ptyinkcode);
+}
+
+VOID changeoutkcode(VOID_A)
+{
+	ptylist[win].outcode = ptyoutkcode;
+	if (!emupid) return;
+
+	sendword(emufd, TE_CHANGEOUTKCODE);
+	sendword(emufd, win);
+	sendword(emufd, ptyoutkcode);
+}
 #endif	/* !_NOKANJICONV */
 
 static int NEAR ptygetkey(VOID_A)
@@ -375,6 +404,9 @@ static int NEAR ptygetkey(VOID_A)
 			kbhit2(1000000L / SENSEPERSEC);
 			waitpty();
 		}
+#ifndef	_NOORIGSHELL
+		if (isshptymode()) break;
+#endif
 		if (c < 0 || ptymenukey < 0 || c != ptymenukey) break;
 
 		str[0] = asprintf3(PTYAI_K, getkeysym(ptymenukey, 0));
@@ -588,7 +620,7 @@ int w;
 	keyseq_t key;
 	char *cp, *func1, *func2, ***varp, **var;
 	u_char uc;
-	int n, fd, val;
+	int n, fd, val, wastty;
 
 	fd = ptylist[w].pipe;
 	if (recvbuf(fd, &uc, sizeof(uc)) < 0) return;
@@ -821,6 +853,44 @@ int w;
 			sendword(emufd, TE_UNLOCKBACK);
 			sendword(emufd, w);
 			break;
+		case TE_SAVETTYIO:
+			if (recvbuf(fd, &n, sizeof(n)) < 0
+			|| recvbuf(fd, &val, sizeof(val)) < 0)
+				break;
+			if (!val) cp = NULL;
+			else {
+				cp = malloc2(val);
+				if (recvbuf(fd, cp, val) < 0) {
+					free(cp);
+					break;
+				}
+			}
+			if (n >= 0) {
+				if (duptty[n]) free(duptty[n]);
+				duptty[n] = cp;
+			}
+			else if (cp) free(cp);
+			break;
+		case TE_INTERNAL:
+			if (recvbuf(fd, &n, sizeof(n)) < 0
+			|| recvstring(fd, &cp) < 0)
+				break;
+			if (n < 0 || n >= FUNCLISTSIZ) {
+				if (cp) free(cp);
+				break;
+			}
+			keywaitfunc = lastfunc;
+			changewin(MAXWINDOWS, (p_id_t)-1);
+			if (!(wastty = isttyiomode)) Xttyiomode(0);
+			ptyinternal++;
+			internal_status = (*funclist[n].func)(cp);
+			ptyinternal--;
+			if (!wastty) Xstdiomode();
+			changewin(win, (p_id_t)-1);
+			rewritefile(0);
+			keywaitfunc = waitpty;
+			if (cp) free(cp);
+			break;
 		case TE_CHANGESTATUS:
 			if (recvbuf(fd, &n, sizeof(n)) < 0) break;
 			ptylist[w].status = n;
@@ -841,6 +911,7 @@ int frontend(VOID_A)
 	changewsize(wheader, windows);
 	changewin(win, ptylist[win].pid);
 
+	internal_status = FNC_FAIL;
 	if (!(wastty = isttyiomode)) Xttyiomode(1);
 	while ((ch = ptygetkey()) >= 0) sendword(emufd, ch);
 	if (!wastty) Xstdiomode();
@@ -856,7 +927,8 @@ int frontend(VOID_A)
 			recvchild(win);
 			killpty(win, &status);
 		}
-		internal_status = (status < 128) ? status : FNC_FAIL;
+		if (internal_status == FNC_FAIL && status < 128)
+			internal_status = status;
 	}
 
 	sigvecset(n);
