@@ -65,7 +65,7 @@ extern int errno;
 #define	getconstvar(s)		(char *)getenv(s)
 #define	getshellvar(s, l)	(char *)getenv(s)
 #else
-#define	getconstvar(s)		(getshellvar(s, sizeof(s) - 1))
+#define	getconstvar(s)		(getshellvar(s, strsize(s)))
 #endif
 
 #ifdef	USESTRERROR
@@ -79,8 +79,15 @@ extern char *sys_errlist[];
 
 #ifdef	_NOPTY
 #define	isptymode()		0
+#define	isshptymode()		0
 #else	/* !_NOPTY */
-#define	isptymode()		ptymode
+# ifdef	_NOORIGSHELL
+# define	isptymode()	ptymode
+# define	isshptymode()	0
+# else
+# define	isptymode()	(ptymode && (fdmode || !shptymode))
+# define	isshptymode()	(!fdmode && shptymode)
+# endif
 #endif	/* !_NOPTY */
 
 /* main.c */
@@ -110,6 +117,23 @@ extern FILE *dopopen __P_((char *));
 extern int dopclose __P_((FILE *));
 #endif
 
+/* log.c */
+#ifdef	_NOLOGGING
+#define	LOG0(l, n, f)
+#define	LOG1(l, n, f, a1)
+#define	LOG2(l, n, f, a1, a2)
+#define	LOG3(l, n, f, a1, a2, a3)
+#else	/* !_NOLOGGING */
+#define	LOG0(l, n, f)		logsyscall(l, n, f)
+#define	LOG1(l, n, f, a1)	logsyscall(l, n, f, a1)
+#define	LOG2(l, n, f, a1, a2)	logsyscall(l, n, f, a1, a2)
+#define	LOG3(l, n, f, a1, a2, a3) \
+				logsyscall(l, n, f, a1, a2, a3)
+extern VOID logclose __P_((VOID_A));
+extern VOID logsyscall __P_((int, int, CONST char *, ...));
+extern VOID logmessage __P_((int, CONST char *, ...));
+#endif	/* !_NOLOGGING */
+
 /* dosemu.c or unixemu.c */
 #ifdef	_USEDOSPATH
 extern int _dospath __P_((char *));
@@ -243,6 +267,8 @@ extern int ptymacro __P_((char *, char *, int));
 #define	ptysystem(c)		ptymacro(c, NULL, F_DOSYSTEM)
 extern VOID killpty __P_((int, int *));
 extern VOID killallpty __P_((VOID_A));
+extern int checkpty __P_((int));
+extern int checkallpty __P_((VOID_A));
 #endif	/* !_NOPTY */
 
 /* frontend.c */
@@ -285,6 +311,8 @@ extern VOID insertwin __P_((int, int));
 extern VOID deletewin __P_((int, int));
 # ifndef	_NOKANJICONV
 extern VOID changekcode __P_((VOID_A));
+extern VOID changeinkcode __P_((VOID_A));
+extern VOID changeoutkcode __P_((VOID_A));
 # endif
 extern int frontend __P_((VOID_A));
 #endif	/* !_NOPTY */
@@ -352,9 +380,7 @@ extern int cmptree __P_((CONST VOID_P, CONST VOID_P));
 extern struct dirent *searchdir __P_((DIR *, reg_t *, char *));
 extern int underhome __P_((char *));
 extern int preparedir __P_((char *));
-#ifdef	NOFLOCK
-#define	lockfile(f, m)
-#else
+#ifndef	NOFLOCK
 extern int lockfile __P_((int, int));
 #endif
 extern int touchfile __P_((char *, struct stat *));
@@ -474,6 +500,9 @@ extern int parsesetdrv __P_((int, char *[], devinfo *));
 extern int parsekeymap __P_((int, char *[], keyseq_t *));
 extern VOID printkeymap __P_((keyseq_t *, int, FILE *));
 #endif
+#if	!MSDOS
+extern int savestdio __P_((int));
+#endif
 #ifdef	_NOORIGSHELL
 extern int searchalias __P_((char *, int));
 extern int addalias __P_((char *, char *));
@@ -521,6 +550,7 @@ extern VOID freedefine __P_((VOID_A));
 #define	BL_KCONV	"kconv"
 #define	BL_READLINE	"readline"
 #define	BL_YESNO	"yesno"
+#define	BL_SAVETTY	"savetty"
 #define	BL_ALIAS	"alias"
 #define	BL_UALIAS	"unalias"
 #define	BL_FUNCTION	"function"
@@ -551,7 +581,7 @@ extern char *removehist __P_((int));
 extern int loadhistory __P_((int, char *));
 extern VOID convhistory __P_((char *, FILE *));
 extern int savehistory __P_((int, char *));
-extern int parsehist __P_((char *, int *));
+extern int parsehist __P_((char *, int *, int));
 extern char *evalhistory __P_((char *));
 #ifdef	DEBUG
 extern VOID freehistory __P_((int));
@@ -586,6 +616,9 @@ extern char *newkanjiconv __P_((char *, int, int, int));
 #ifndef	_NOKANJIFCONV
 extern int getkcode __P_((char *));
 #endif
+#ifndef	_NOKANJICONV
+extern int getoutputkcode __P_((VOID_A));
+#endif
 extern char *convget __P_((char *, char *, int));
 extern char *convput __P_((char *, char *, int, int, char *, int *));
 
@@ -593,6 +626,8 @@ extern char *convput __P_((char *, char *, int, int, char *, int *));
 extern int intrkey __P_((VOID_A));
 extern int Xgetkey __P_((int, int));
 extern int kanjiputs2 __P_((char *, int, int));
+extern VOID cputspace __P_((int));
+extern VOID cputstr __P_((int, char *));
 extern char *inputstr __P_((char *, int, int, char *, int));
 extern int yesno __P_((CONST char *, ...));
 extern VOID warning __P_((int, char *));
