@@ -28,27 +28,27 @@ extern int errno;
 
 #ifdef	USEMNTENTH
 #include <mntent.h>
-typedef struct mntent	mnt_t;
-#define	setmntent2	setmntent
-#define	getmntent2(f,m)	getmntent(f)
-#define	hasmntopt2(m,o)	strmntopt((m) -> mnt_opts, o)
-#define	endmntent2	endmntent
+typedef struct mntent		mnt_t;
+#define	setmntent2		setmntent
+#define	getmntent2(f,m)		getmntent(f)
+#define	hasmntopt2(m,o)		strmntopt((m) -> mnt_opts, o)
+#define	endmntent2		endmntent
 #endif	/* USEMNTENTH */
 
 #ifdef	USEMNTTABH
 #include <sys/mnttab.h>
-#define	MOUNTED		MNTTAB
-typedef struct mnttab	mnt_t;
-#define	setmntent2	fopen
-#define	getmntent2(f,m)	(getmntent(f, m) ? NULL : m)
-#define	hasmntopt2(m,o)	strmntopt((m) -> mnt_mntopts, o)
-#define	endmntent2	fclose
-#define	mnt_dir		mnt_mountp
-#define	mnt_fsname	mnt_special
-#define	mnt_type	mnt_fstype
-#ifndef	mnt_opts
-#define	mnt_opts	mnt_mntopts
-#endif
+#define	MOUNTED			MNTTAB
+typedef struct mnttab		mnt_t;
+#define	setmntent2		fopen
+#define	getmntent2(f,m)		(getmntent(f, m) ? NULL : m)
+#define	hasmntopt2(m,o)		strmntopt((m) -> mnt_mntopts, o)
+#define	endmntent2		fclose
+#define	mnt_dir			mnt_mountp
+#define	mnt_fsname		mnt_special
+#define	mnt_type		mnt_fstype
+# ifndef	mnt_opts
+# define	mnt_opts	mnt_mntopts
+# endif
 #endif	/* USEMNTTABH */
 
 #ifdef	USEMNTCTL
@@ -59,7 +59,8 @@ typedef struct mnttab	mnt_t;
 #endif	/* USEMNTCTL */
 
 #if	!defined (USEMOUNTH) && !defined (USEFSDATA) \
-&& (defined (USEGETFSSTAT) || defined (USEMNTINFOR) || defined (USEMNTINFO))
+&& (defined (USEGETFSSTAT) || defined (USEGETVFSTAT) \
+|| defined (USEMNTINFOR) || defined (USEMNTINFO))
 #include <sys/mount.h>
 #endif
 
@@ -67,8 +68,9 @@ typedef struct mnttab	mnt_t;
 #include <sys/fs_types.h>
 #endif
 
-#if	defined (USEGETFSSTAT) || defined (USEMNTCTL) \
-|| defined (USEMNTINFOR) || defined (USEMNTINFO) || defined (USEGETMNT)
+#if	defined (USEGETFSSTAT) || defined (USEGETVFSTAT) \
+|| defined (USEMNTCTL) || defined (USEMNTINFOR) || defined (USEMNTINFO) \
+|| defined (USEGETMNT)
 typedef struct _mnt_t {
 	char *mnt_fsname;
 	char *mnt_dir;
@@ -77,26 +79,27 @@ typedef struct _mnt_t {
 } mnt_t;
 static FILE *NEAR setmntent2 __P_((char *, char *));
 static mnt_t *NEAR getmntent2 __P_((FILE *, mnt_t *));
-#define	hasmntopt2(mntp, opt)	strmntopt((mntp) -> mnt_opts, opt)
-#if	defined (USEMNTINFO) || defined (USEGETMNT)
-#define	endmntent2(f)
-#else
-#define	endmntent2		free
-#endif
+#define	hasmntopt2(m,o)		strmntopt((m) -> mnt_opts, o)
+# if	defined (USEMNTINFO) || defined (USEGETMNT)
+# define	endmntent2(f)
+# else
+# define	endmntent2	free
+# endif
 static int mnt_ptr = 0;
 static int mnt_size = 0;
-#endif
+#endif	/* USEGETFSSTAT || USEGETVFSTAT || USEMNTCTL \
+|| USEMNTINFOR || USEMNTINFO || USEGETMNT */
 
 #ifdef	USEGETFSENT
 #include <fstab.h>
-typedef struct fstab	mnt_t;
-#define	setmntent2(file, mode)	(FILE *)(setfsent(), NULL)
-#define	getmntent2(fp, mntp)	getfsent()
-#define	hasmntopt2(mntp, opt)	strmntopt((mntp) -> fs_mntops, opt)
+typedef struct fstab		mnt_t;
+#define	setmntent2(f,m)		(FILE *)(setfsent(), NULL)
+#define	getmntent2(f,m)		getfsent()
+#define	hasmntopt2(m,o)		strmntopt((m) -> fs_mntops, o)
 #define	endmntent2(fp)		endfsent()
-#define	mnt_dir		fs_file
-#define	mnt_fsname	fs_spec
-#define	mnt_type	fs_vfstype
+#define	mnt_dir			fs_file
+#define	mnt_fsname		fs_spec
+#define	mnt_type		fs_vfstype
 #endif	/* USEGETFSENT */
 
 #if	MSDOS
@@ -106,7 +109,7 @@ typedef struct _mnt_t {
 	char *mnt_type;
 	char *mnt_opts;
 } mnt_t;
-#define	hasmntopt2(mntp, opt)	strmntopt((mntp) -> mnt_opts, opt)
+#define	hasmntopt2(m,o)		strmntopt((m) -> mnt_opts, o)
 # ifdef	PC98
 # define	PT_FAT12	0x81	/* 0x80 | 0x01 */
 # define	PT_FAT16	0x91	/* 0x80 | 0x11 */
@@ -121,51 +124,53 @@ typedef struct _mnt_t {
 # endif
 #endif	/* MSDOS */
 
+#if	defined (USESTATVFSH) || defined (USEGETVFSTAT)
+#include <sys/statvfs.h>
+#endif
 
 #ifdef	USESTATVFSH
-#include <sys/statvfs.h>
 # ifdef	USESTATVFS_T
-typedef statvfs_t	statfs_t;
+typedef statvfs_t		statfs_t;
 # else
-typedef struct statvfs	statfs_t;
+typedef struct statvfs		statfs_t;
 # endif
-#define	statfs2		statvfs
-#define	blocksize(fs)	((fs).f_frsize ? (fs).f_frsize : (fs).f_bsize)
+#define	statfs2			statvfs
+#define	blocksize(fs)		((fs).f_frsize ? (fs).f_frsize : (fs).f_bsize)
 #endif	/* USESTATVFSH */
 
 #ifdef	USESTATFSH
 #include <sys/statfs.h>
-#define	f_bavail	f_bfree
-typedef struct statfs	statfs_t;
-#define	blocksize(fs)	(fs).f_bsize
+#define	f_bavail		f_bfree
+typedef struct statfs		statfs_t;
+#define	blocksize(fs)		(fs).f_bsize
 #endif	/* USESTATFSH */
 
 #ifdef	USEVFSH
 #include <sys/vfs.h>
-typedef struct statfs	statfs_t;
-#define	blocksize(fs)	(fs).f_bsize
+typedef struct statfs		statfs_t;
+#define	blocksize(fs)		(fs).f_bsize
 #endif	/* USEVFSH */
 
 #ifdef	USEMOUNTH
 #include <sys/mount.h>
-typedef struct statfs	statfs_t;
-#define	blocksize(fs)	(fs).f_bsize
+typedef struct statfs		statfs_t;
+#define	blocksize(fs)		(fs).f_bsize
 #endif	/* USEMOUNTH */
 
 #ifdef	USEFSDATA
 #include <sys/mount.h>
-typedef struct fs_data	statfs_t;
+typedef struct fs_data		statfs_t;
 #define	statfs2(path, buf)	(statfs(path, buf) - 1)
 #define	blocksize(fs)		1024
-#define	f_bsize		fd_req.bsize
-#define	f_files		fd_req.gtot
-#define	f_blocks	fd_req.btot
-#define	f_bfree		fd_req.bfree
-#define	f_bavail	fd_req.bfreen
+#define	f_bsize			fd_req.bsize
+#define	f_files			fd_req.gtot
+#define	f_blocks		fd_req.btot
+#define	f_bfree			fd_req.bfree
+#define	f_bavail		fd_req.bfreen
 #endif	/* USEFSDATA */
 
 #ifdef	USEFFSIZE
-#define	f_bsize		f_fsize
+#define	f_bsize			f_fsize
 #endif
 
 #if	defined (USESTATFSH) || defined (USEVFSH) || defined (USEMOUNTH)
@@ -189,8 +194,8 @@ typedef struct _statfs_t {
 	long f_files;
 } statfs_t;
 extern int unixstatfs __P_((char *, statfs_t *));
-#define	statfs2		unixstatfs
-#define	blocksize(fs)	(fs).f_bsize
+#define	statfs2			unixstatfs
+#define	blocksize(fs)		(fs).f_bsize
 #endif
 
 extern VOID error __P_((char *));
@@ -217,6 +222,8 @@ extern int Xstat __P_((char *, struct stat *));
 #endif
 extern int Xaccess __P_((char *, int));
 extern int filetop __P_((int));
+extern VOID cputspace __P_((int));
+extern VOID cputstr __P_((int, char *));
 #ifdef	_NOPTY
 #define	Xlocate			locate
 #define	Xputterm		putterm
@@ -329,7 +336,7 @@ static CONST int keycodelist[] = {
 	K_BEG, K_EOL, K_NPAGE, K_PPAGE, K_CLR, K_ENTER, K_HELP,
 	K_BS, '\t', K_CR, K_ESC
 };
-#define	KEYCODESIZ	((int)(sizeof(keycodelist) / sizeof(int)))
+#define	KEYCODESIZ	arraysize(keycodelist)
 static CONST char *keystrlist[] = {
 	"Home", "End", "DelLin", "InsLin", "Del", "Ins",
 	"Beg", "Eol", "PageDn", "PageUp", "Clr", "Enter", "Help",
@@ -365,13 +372,13 @@ static CONST strtable mntlist[] = {
 	{FSID_LFN, MNTTYPE_VFAT},
 	{0, MNTTYPE_ADVFS},
 	{0, MNTTYPE_VXFS},
-#endif	/* !MSDOS */
-#ifdef	DARWIN
+# ifdef	DARWIN
 	/* Macintosh HFS+ is pseudo file system covered with skin */
 	{0, MNTTYPE_HFS},
-#endif
+# endif
+#endif	/* !MSDOS */
 };
-#define	MNTLISTSIZ	((int)(sizeof(mntlist) / sizeof(strtable)))
+#define	MNTLISTSIZ	arraysize(mntlist)
 
 
 static int NEAR code2str(buf, code)
@@ -475,7 +482,7 @@ int mode;
 		if (!c) continue;
 
 		Xcputs2("  ");
-		if (c < 2) Xcprintf2("%*s", KEYWID, "");
+		if (c < 2) cputspace(KEYWID);
 		Xcprintf2("%k: %k",
 			buf, mesconv(funclist[i].hmes, funclist[i].hmes_eng));
 
@@ -556,85 +563,111 @@ mnt_t *mntp;
 	mntp -> mnt_fsname = fsname;
 	mntp -> mnt_dir = dir;
 	mntp -> mnt_type = (type) ? type : "???";
-	mntp -> mnt_opts = (vmntp -> vmt_flags & MNT_READONLY) ? "ro" : "";
+	mntp -> mnt_opts =
+		(vmntp -> vmt_flags & MNT_READONLY) ? "ro" : nullstr;
 	mnt_ptr += vmntp -> vmt_length;
 
 	return(mntp);
 }
 #endif	/* USEMNTCTL */
 
-#if	defined (USEMNTINFOR) || defined (USEMNTINFO) || defined (USEGETFSSTAT)
+#if	defined (USEMNTINFOR) || defined (USEMNTINFO) \
+|| defined (USEGETFSSTAT) || defined (USEGETVFSTAT)
+
+#ifdef	USEGETVFSTAT
+#define	f_flags			f_flag
+#define	getfsstat2		getvfsstat
+typedef struct statvfs		mntinfo_t;
+#else
+#define	getfsstat2		getfsstat
+typedef struct statfs		mntinfo_t;
+#endif
+
+#if	!defined (MNT_RDONLY) && defined (M_RDONLY)
+#define	MNT_RDONLY		M_RDONLY
+#endif
+
 /*ARGSUSED*/
 static FILE *NEAR setmntent2(file, mode)
 char *file, *mode;
 {
-#ifndef	USEMNTINFO
+# ifndef	USEMNTINFO
 	int size;
-#endif
-	struct statfs *buf;
+# endif
+	mntinfo_t *buf;
 
 	buf = NULL;
 	mnt_ptr = mnt_size = 0;
-#ifdef	USEMNTINFO
+
+# ifdef	DEBUG
+	_mtrace_file = "getmntinfo(start)";
+# endif
+# ifdef	USEMNTINFO
 	mnt_size = getmntinfo(&buf, MNT_NOWAIT);
-#else
-# ifdef	USEMNTINFOR
+# else	/* !USEMNTINFO */
+#  ifdef	USEMNTINFOR
 	size = 0;
 	getmntinfo_r(&buf, MNT_WAIT, &mnt_size, &size);
-# else
-	size = (getfsstat(NULL, 0, MNT_WAIT) + 1) * sizeof(struct statfs);
+#  else
+	size = (getfsstat2(NULL, 0, MNT_WAIT) + 1) * sizeof(mntinfo_t);
 	if (size > 0) {
-		buf = (struct statfs *)malloc2(mnt_size);
-		mnt_size = getfsstat(buf, mnt_size, MNT_WAIT);
+		buf = (mntinfo_t *)malloc2(size);
+		mnt_size = getfsstat2(buf, size, MNT_WAIT);
+	}
+#  endif
+# endif	/* !USEMNTINFO */
+# ifdef	DEBUG
+	if (_mtrace_file) _mtrace_file = NULL;
+	else {
+		_mtrace_file = "getmntinfo(end)";
+		malloc(0);	/* dummy malloc */
 	}
 # endif
-#endif
 
 	return((FILE *)buf);
 }
 
-#if	!defined (MNT_RDONLY) && defined (M_RDONLY)
-#define	MNT_RDONLY	M_RDONLY
-#endif
-
 static mnt_t *NEAR getmntent2(fp, mntp)
 FILE *fp;
 mnt_t *mntp;
 {
-#ifdef	USEMNTINFO
-# ifdef	USEVFCNAME
+# if	defined (USEMNTINFO) || defined (USEGETVFSTAT)
+#  ifdef	USEVFCNAME
 	struct vfsconf *conf;
-# define	getvfsbynumber(n)	((conf = getvfsbytype(n)) \
+#  define	getvfsbynumber(n)	((conf = getvfsbytype(n)) \
 					? conf -> vfc_name : NULL)
-# else	/* !USEVFCNAME */
-#  ifdef	USEFFSTYPE
-#  define	getvfsbynumber(n)	(buf[mnt_ptr].f_fstypename)
-#  else
-#   ifdef	INITMOUNTNAMES
+#  else	/* !USEVFCNAME */
+#   ifdef	USEFFSTYPE
+#   define	getvfsbynumber(n)	(buf[mnt_ptr].f_fstypename)
+#   else	/* !USEFFSTYPE */
+#    ifdef	INITMOUNTNAMES
 	static char *mnt_names[] = INITMOUNTNAMES;
-#   define	getvfsbynumber(n)	(((n) <= MOUNT_MAXTYPE) \
+#    define	getvfsbynumber(n)	(((n) <= MOUNT_MAXTYPE) \
+					? mnt_names[n] : NULL)
+#    else
+#    define	getvfsbynumber(n)	(NULL)
+#    endif
+#   endif	/* !USEFFSTYPE */
+#  endif	/* !USEVFCNAME */
+# else	/* !USEMNTINFO && !USEGETVFSTAT */
+#  ifdef	USEGETFSSTAT
+#  define	getvfsbynumber(n)	(((n) <= MOUNT_MAXTYPE) \
 					? mnt_names[n] : NULL)
-#   else
-#   define	getvfsbynumber(n)	(NULL)
-#   endif
 #  endif
-# endif	/* !USEVFCNAME */
-#else	/* !USEMNTINFO */
-# ifdef	USEGETFSSTAT
-# define	getvfsbynumber(n)	(((n) <= MOUNT_MAXTYPE) \
-					? mnt_names[n] : NULL)
-# endif
-#endif	/* !USEMNTINFO */
+# endif	/* !USEMNTINFO && !USEGETVFSTAT */
 	static char *fsname = NULL;
 	static char *dir = NULL;
 	static char *type = NULL;
-	struct statfs *buf;
+	mntinfo_t *buf;
 	char *cp;
 	ALLOC_T len;
 
 	if (mnt_ptr >= mnt_size) return(NULL);
-	buf = (struct statfs *)fp;
+	buf = (mntinfo_t *)fp;
 
+# ifdef	DEBUG
+	_mtrace_file = "getmntent(start)";
+# endif
 	len = strlen(buf[mnt_ptr].f_mntfromname) + 1;
 	fsname = realloc2(fsname, len);
 	memcpy(fsname, buf[mnt_ptr].f_mntfromname, len);
@@ -653,16 +686,24 @@ mnt_t *mntp;
 		free(type);
 		type = NULL;
 	}
+# ifdef	DEBUG
+	if (_mtrace_file) _mtrace_file = NULL;
+	else {
+		_mtrace_file = "getmntent(end)";
+		malloc(0);	/* dummy malloc */
+	}
+# endif
 
 	mntp -> mnt_fsname = fsname;
 	mntp -> mnt_dir = dir;
 	mntp -> mnt_type = (type) ? type : "???";
-	mntp -> mnt_opts = (buf[mnt_ptr].f_flags & MNT_RDONLY) ? "ro" : "";
+	mntp -> mnt_opts =
+		(buf[mnt_ptr].f_flags & MNT_RDONLY) ? "ro" : nullstr;
 	mnt_ptr++;
 
 	return(mntp);
 }
-#endif	/* USEMNTINFOR || USEMNTINFO || USEGETFSSTAT */
+#endif	/* USEMNTINFOR || USEMNTINFO || USEGETFSSTAT || USEGETVFSTAT */
 
 #ifdef	USEGETMNT
 /*ARGSUSED*/
@@ -703,7 +744,7 @@ mnt_t *mntp;
 	mntp -> mnt_fsname = fsname;
 	mntp -> mnt_dir = dir;
 	mntp -> mnt_type = type;
-	mntp -> mnt_opts = (buf.fd_req.flags & M_RONLY) ? "ro" : "";
+	mntp -> mnt_opts = (buf.fd_req.flags & M_RONLY) ? "ro" : nullstr;
 
 	return(mntp);
 }
@@ -765,7 +806,7 @@ mnt_t *mntbuf;
 			break;
 	}
 # endif	/* !_NOUSELFN */
-	mntbuf -> mnt_opts = "";
+	mntbuf -> mnt_opts = nullstr;
 
 	if (statfs2(mntbuf -> mnt_dir, fsbuf) < 0) return(-1);
 #else	/* !MSDOS */
@@ -781,38 +822,38 @@ mnt_t *mntbuf;
 	if (!mntbuf) mntbuf = &mnt;
 
 	dir = NULL;
-	if (!strncmp(path, "/dev/", sizeof("/dev/") - 1)) {
+	if (!strncmp(path, "/dev/", strsize("/dev/"))) {
 		if (_chdir2(path) < 0) dir = path;
 	}
 # ifndef	_NODOSDRIVE
 	else if ((drv = dospath(path, NULL))) {
 		static char dosmntdir[4];
-		char buf[sizeof(long) * 3 + 1];
+		char buf[3 * sizeof(long) + 1];
 
-		mntbuf -> mnt_fsname = "";
+		mntbuf -> mnt_fsname = nullstr;
 		mntbuf -> mnt_dir = dosmntdir;
 		dosmntdir[0] = drv;
 		strcpy(&(dosmntdir[1]), ":\\");
 		mntbuf -> mnt_type =
 			(islower2(drv)) ? MNTTYPE_DOS7 : MNTTYPE_PC;
-		mntbuf -> mnt_opts = "";
+		mntbuf -> mnt_opts = nullstr;
 		if (dosstatfs(drv, buf) < 0) return(-1);
-		if (buf[sizeof(long) * 3] & 001)
+		if (buf[3 * sizeof(long)] & 001)
 			mntbuf -> mnt_type = MNTTYPE_FAT32;
-		fsbuf -> f_bsize = *((long *)&(buf[sizeof(long) * 0]));
+		fsbuf -> f_bsize = *((long *)&(buf[0 * sizeof(long)]));
 #  ifdef	USEFSDATA
 		fsbuf -> fd_req.btot = calcKB((off_t)(fsbuf -> f_bsize),
-			(off_t)(*((long *)&(buf[sizeof(long) * 1]))));
+			(off_t)(*((long *)&(buf[1 * sizeof(long)]))));
 		fsbuf -> fd_req.bfree =
 		fsbuf -> fd_req.bfreen = calcKB((off_t)(fsbuf -> f_bsize),
-			(off_t)(*((long *)&(buf[sizeof(long) * 2]))));
+			(off_t)(*((long *)&(buf[2 * sizeof(long)]))));
 #  else	/* !USEFSDATA */
 #   ifdef	USESTATVFSH
 		fsbuf -> f_frsize = 0L;
 #   endif
-		fsbuf -> f_blocks = *((long *)&(buf[sizeof(long) * 1]));
+		fsbuf -> f_blocks = *((long *)&(buf[1 * sizeof(long)]));
 		fsbuf -> f_bfree =
-		fsbuf -> f_bavail = *((long *)&(buf[sizeof(long) * 2]));
+		fsbuf -> f_bavail = *((long *)&(buf[2 * sizeof(long)]));
 #  endif	/* !USEFSDATA */
 		fsbuf -> f_files = -1L;
 
@@ -931,7 +972,7 @@ char *dir;
 	statfs_t fsbuf;
 	struct stat st;
 
-	if (strcmp(dir, ".")) {
+	if (isdotdir(dir) != 2) {
 		/* In case of the newborn directory */
 		if (Xstat(dir, &st) >= 0) return((off_t)(st.st_size));
 	}
@@ -968,7 +1009,7 @@ char *s, *unit;
 	Xlocate(n_column / 2 + 2, yy + y);
 	if (s) {
 		width = n_column - (n_column / 2 + 2);
-		Xcprintf2("%-*.*k", width, width, s);
+		cputstr(width, s);
 	}
 	else Xcprintf2("%<'*qd %k", MAXCOLSCOMMA(3), n, unit);
 
