@@ -1,14 +1,12 @@
 /*
  *	libc.c
  *
- *	Arrangememt of Library Function
+ *	arrangememt of library functions
  */
 
 #include <fcntl.h>
 #include "fd.h"
-#include "term.h"
 #include "func.h"
-#include "kctype.h"
 #include "kanji.h"
 
 #ifndef	NOTZFILEH
@@ -21,18 +19,16 @@ extern int setcurdrv __P_((int, int));
 extern char *unixrealpath __P_((char *, char *));
 #endif
 
-#ifndef	_NODOSDRIVE
-extern int flushdrv __P_((int, VOID_T (*)__P_((VOID_A))));
-extern char *dosgetcwd __P_((char *, int));
-#endif
-
 #ifdef	_NOORIGSHELL
-#include <signal.h>
-#define	dosystem(s)		system(s)
+#include "wait.h"
 #else
 #include "system.h"
 #endif
 
+#ifndef	_NOPTY
+#include "termemu.h"
+#endif
+
 #ifdef	CYGWIN
 #include <sys/cygwin.h>
 #endif
@@ -51,21 +47,6 @@ extern int hideclock;
 extern int lastdrv;
 #endif
 
-#ifndef	SIG_ERR
-#define	SIG_ERR		((sigcst_t)-1)
-#endif
-#ifndef	CHAR_BIT
-# ifdef	NBBY
-# define	CHAR_BIT	NBBY
-# else
-# define	CHAR_BIT	0x8
-# endif
-#endif
-#define	char2long(cp)	(  ((u_char *)(cp))[3] \
-			| (((u_char *)(cp))[2] << (CHAR_BIT * 1)) \
-			| (((u_char *)(cp))[1] << (CHAR_BIT * 2)) \
-			| (((u_char *)(cp))[0] << (CHAR_BIT * 3)) )
-
 #ifndef	NOSYMLINK
 static int NEAR evallink __P_((char *, char *));
 #endif
@@ -76,9 +57,11 @@ static char **NEAR _putenv2 __P_((char *, char **));
 #endif
 #if	!MSDOS && !defined (NOTZFILEH) \
 && !defined (USEMKTIME) && !defined (USETIMELOCAL)
+static long NEAR char2long __P_((u_char *));
 static int NEAR tmcmp __P_((struct tm *, struct tm *));
 #endif
 #if	!defined (USEMKTIME) && !defined (USETIMELOCAL)
+static time_t NEAR timegm2 __P_((struct tm *));
 static long NEAR gettimezone __P_((struct tm *, time_t));
 #endif
 
@@ -93,8 +76,6 @@ static char *lastpath = NULL;
 static char *unixpath = NULL;
 #endif
 static int wasttyflags = 0;
-#define	TF_TTYIOMODE	001
-#define	TF_TTYNL	002
 
 
 int stat2(path, stp)
@@ -114,13 +95,13 @@ struct stat *stp;
 			return(-1);
 		}
 # endif
-		if (nodoslstat(path, stp) < 0
-		|| (stp -> st_mode & S_IFMT) != S_IFLNK) {
+		if (nodoslstat(path, stp) < 0 || !s_islnk(stp)) {
 			errno = duperrno;
 			return(-1);
 		}
 		stp -> st_mode &= ~S_IFMT;
 	}
+
 	return(0);
 #endif	/* !NOSYMLINK */
 }
@@ -138,7 +119,7 @@ char *path, *delim;
 		errno = duperrno;
 		return(-1);
 	}
-	if ((st.st_mode & S_IFMT) != S_IFLNK) {
+	if (!s_islnk(&st)) {
 		errno = duperrno;
 		return(0);
 	}
@@ -154,6 +135,7 @@ char *path, *delim;
 	buf[i] = *delim = '\0';
 	_realpath2(buf, path, 1);
 	errno = duperrno;
+
 	return(1);
 }
 #endif	/* !NOSYMLINK */
@@ -164,8 +146,9 @@ char *path, *resolved;
 int rdlink;
 {
 	char *cp, *top;
+	int n;
 
-	if (!*path || !strcmp(path, ".")) return(resolved);
+	if (!*path || (n = isdotdir(path)) == 2) return(resolved);
 	else if ((cp = strdelim(path, 0))) {
 		*cp = '\0';
 		_realpath2(path, resolved, rdlink);
@@ -174,7 +157,7 @@ int rdlink;
 		return(resolved);
 	}
 
-	if (!strcmp(path, "..")) {
+	if (n == 1) {
 		cp = strrdelim(resolved, 0);
 		top = resolved;
 #if	MSDOS
@@ -204,6 +187,7 @@ int rdlink;
 		else evallink(resolved, cp - 1);
 #endif	/* !NOSYMLINK */
 	}
+
 	return(resolved);
 }
 
@@ -224,7 +208,7 @@ int rdlink;
 	path = tmp;
 
 #if	MSDOS
-	drv = dospath("", NULL);
+	drv = dospath(nullstr, NULL);
 	if ((resolved[0] = _dospath(path))) path += 2;
 	if (*path == _SC_) {
 		if (!resolved[0]) resolved[0] = drv;
@@ -244,7 +228,7 @@ int rdlink;
 		}
 	}
 #else	/* !MSDOS */
-	if (*path == _SC_) strcpy(resolved, _SS_);
+	if (*path == _SC_) copyrootpath(resolved);
 # ifndef	_NODOSDRIVE
 	else if ((drv = _dospath(path))) {
 		path += 2;
@@ -266,10 +250,11 @@ int rdlink;
 #endif	/* !MSDOS */
 	else if (!rdlink && resolved != fullpath && *fullpath)
 		strcpy(resolved, fullpath);
-	else if (!Xgetwd(resolved)) strcpy(resolved, _SS_);
+	else if (!Xgetwd(resolved)) copyrootpath(resolved);
 	norealpath++;
 	_realpath2(path, resolved, rdlink);
 	norealpath--;
+
 	return(resolved);
 }
 
@@ -278,17 +263,17 @@ char *path;
 {
 	char cwd[MAXPATHLEN];
 
-	if (!Xgetwd(cwd)) strcpy(cwd, _SS_);
+	if (!Xgetwd(cwd)) copyrootpath(cwd);
 	if (Xchdir(path) < 0) return(-1);
 #if	!MSDOS
 # ifndef	_NODOSDRIVE
-	if (dospath2("")) /*EMPTY*/;
+	if (dospath2(nullstr)) /*EMPTY*/;
 	else
 # endif
 	{
 		int fd, duperrno;
 
-		if ((fd = open(".", O_RDONLY, 0600)) < 0) {
+		if ((fd = open(curpath, O_RDONLY, 0600)) < 0) {
 # ifdef	CYGWIN
 			char upath[MAXPATHLEN], spath[MAXPATHLEN];
 			char tmp[MAXPATHLEN];
@@ -313,6 +298,7 @@ char *path;
 		close(fd);
 	}
 #endif	/* !MSDOS */
+
 	return(0);
 }
 
@@ -332,7 +318,7 @@ char *path;
 # endif
 		return(0);
 	}
-#endif
+#endif	/* DEBUG */
 
 	realpath2(path, tmp, physical_path);
 	strcpy(cwd, fullpath);
@@ -366,11 +352,16 @@ char *path;
 	if (unixrealpath(fullpath, tmp)) strcpy(fullpath, tmp);
 #endif
 	entryhist(1, fullpath, 1);
+#ifndef	_NOPTY
+	sendparent(TE_CHDIR, fullpath);
+#endif
+
 	return(0);
 }
 
-int chdir3(path)
+int chdir3(path, raw)
 char *path;
+int raw;
 {
 #ifndef	_NODOSDRIVE
 	int drive;
@@ -378,23 +369,33 @@ char *path;
 	char *cwd;
 
 	cwd = path;
-	if (!strcmp(path, ".")) cwd = NULL;
-	else if (!strcmp(path, "?")) path = origpath;
-	else if (!strcmp(path, "-")) {
-		if (!lastpath) {
-			errno = ENOENT;
-			return(-1);
-		}
-		path = lastpath;
+	if (!raw && path[0] && !path[1]) switch(path[0]) {
+		case '.':
+			cwd = NULL;
+			break;
+		case '?':
+			path = origpath;
+			break;
+		case '-':
+			if (!lastpath) {
+				errno = ENOENT;
+				return(-1);
+			}
+			path = lastpath;
+			break;
+#ifndef	_NODOSDRIVE
+		case '@':
+			if (!unixpath) {
+				errno = ENOENT;
+				return(-1);
+			}
+			path = unixpath;
+			break;
+#endif
+		default:
+			break;
 	}
 #ifndef	_NODOSDRIVE
-	else if (!strcmp(path, "@")) {
-		if (!unixpath) {
-			errno = ENOENT;
-			return(-1);
-		}
-		path = unixpath;
-	}
 	if ((drive = dospath3(fullpath))) flushdrv(drive, NULL);
 #endif
 	if (chdir2(path) < 0) return(-1);
@@ -404,7 +405,11 @@ char *path;
 	else {
 		if (findpattern) free(findpattern);
 		findpattern = NULL;
+#ifndef	_NOUSEHASH
+		searchhash(NULL, nullstr, nullstr);
+#endif
 	}
+
 	return(0);
 }
 
@@ -439,6 +444,7 @@ int mode;
 		if (Xmkdir(path, mode) < 0 && errno != EEXIST) return(-1);
 		cp1 = cp2;
 	}
+
 	return(0);
 }
 
@@ -453,6 +459,7 @@ ALLOC_T size;
 		tmp = NULL;	/* fake for -Wuninitialized */
 #endif
 	}
+
 	return(tmp);
 }
 
@@ -470,6 +477,7 @@ ALLOC_T size;
 		tmp = NULL;	/* fake for -Wuninitialized */
 #endif
 	}
+
 	return(tmp);
 }
 
@@ -482,6 +490,7 @@ ALLOC_T n, *sizep;
 		return(malloc2(*sizep));
 	}
 	while (n + 1 >= *sizep) *sizep *= 2;
+
 	return(realloc2(ptr, *sizep));
 }
 
@@ -495,6 +504,7 @@ char *s;
 	n = strlen(s);
 	if (!(tmp = (char *)malloc((ALLOC_T)n + 1))) error("malloc()");
 	memcpy(tmp, s, n + 1);
+
 	return(tmp);
 }
 
@@ -510,6 +520,7 @@ int n;
 	if (!(tmp = (char *)malloc((ALLOC_T)i + 1))) error("malloc()");
 	memcpy(tmp, s, i);
 	tmp[i] = '\0';
+
 	return(tmp);
 }
 
@@ -522,7 +533,11 @@ int c;
 	for (i = 0; s[i]; i++) {
 		if (s[i] == c) return(&(s[i]));
 		if (iskanji1(s, i)) i++;
+#ifdef	CODEEUC
+		else if (isekana(s, i)) i++;
+#endif
 	}
+
 	return(NULL);
 }
 
@@ -537,7 +552,11 @@ int c;
 	for (i = 0; s[i]; i++) {
 		if (s[i] == c) cp = &(s[i]);
 		if (iskanji1(s, i)) i++;
+#ifdef	CODEEUC
+		else if (isekana(s, i)) i++;
+#endif
 	}
+
 	return(cp);
 }
 
@@ -548,6 +567,7 @@ char *s1, *s2;
 
 	for (i = 0; s2[i]; i++) s1[i] = s2[i];
 	s1[i] = '\0';
+
 	return(&(s1[i]));
 }
 
@@ -559,6 +579,7 @@ int n;
 
 	for (i = 0; i < n && s2[i]; i++) s1[i] = s2[i];
 	s1[i] = '\0';
+
 	return(&(s1[i]));
 }
 
@@ -570,17 +591,16 @@ int strncpy3(s1, s2, lenp, ptr)
 char *s1, *s2;
 int *lenp, ptr;
 {
-	int i, j, l;
+	int i, j, len;
 
 	for (i = j = 0; i < ptr && s2[j]; i++, j++) {
-#ifdef	CODEEUC
-		if (isekana(s2, j)) j++;
-		else
-#endif
 		if (iskanji1(s2, j)) {
 			i++;
 			j++;
 		}
+#ifdef	CODEEUC
+		else if (isekana(s2, j)) j++;
+#endif
 	}
 	if (!i || i <= ptr) i = 0;
 	else {
@@ -588,57 +608,45 @@ int *lenp, ptr;
 		i = 1;
 	}
 
-	while ((*lenp < 0 || i < *lenp) && s2[j]) {
+	while (i < *lenp && s2[j]) {
+		if (iskanji1(s2, j)) {
+			i += snprintf2(&(s1[i]), *lenp - i + 1,
+				"%.2s", &(s2[j]));
+			j += 2;
+			continue;
+		}
 #ifdef	CODEEUC
-		if (isekana(s2, j)) {
-			if (*lenp >= 0) (*lenp)++;
+		else if (isekana(s2, j)) {
+			(*lenp)++;
 			s1[i++] = s2[j++];
 		}
 #else
-		if (iskna(s2[j])) /*EMPTY*/;
+		else if (isskana(s2, j)) /*EMPTY*/;
 #endif
-		else if (iskanji1(s2, j)) {
-			if (*lenp >= 0 && i >= *lenp - 1) {
-				if (ptr >= 0) s1[i++] = ' ';
-				break;
-			}
-			s1[i++] = s2[j++];
-		}
-		else if (isctl(s2[j])) {
-			s1[i++] = '^';
-			if (*lenp >= 0 && i >= *lenp) break;
-			s1[i++] = ((s2[j++] + '@') & 0x7f);
+		else if (iscntrl2(s2[j])) {
+			i += snprintf2(&(s1[i]), *lenp - i + 1,
+				"^%c", (s2[j++] + '@') & 0x7f);
 			continue;
 		}
 		else if (ismsb(s2[j])) {
-			int c;
-
-			c = s2[j++] & 0xff;
-			s1[i++] = '\\';
-			if (*lenp >= 0 && i >= *lenp) break;
-			s1[i++] = (c / (8 * 8)) + '0';
-			if (*lenp >= 0 && i >= *lenp) break;
-			s1[i++] = ((c % (8 * 8)) / 8) + '0';
-			if (*lenp >= 0 && i >= *lenp) break;
-			s1[i++] = (c % 8) + '0';
+			i += snprintf2(&(s1[i]), *lenp - i + 1,
+				"\\%03o", s2[j++] & 0xff);
 			continue;
 		}
 		s1[i++] = s2[j++];
 	}
 
-	l = i;
-	if (*lenp >= 0 && ptr >= 0) while (i < *lenp) s1[i++] = ' ';
+	len = i;
+	if (ptr >= 0) while (i < *lenp) s1[i++] = ' ';
 	s1[i] = '\0';
-	return(l);
+
+	return(len);
 }
 
 int strlen2(s)
 char *s;
 {
-	int i, len;
-
-	for (i = len = 0; s[i]; i++, len++) if (isctl(s[i])) len++;
-	return(len);
+	return(snprintf2(NULL, 0, "%s", s));
 }
 
 int strlen3(s)
@@ -647,84 +655,52 @@ char *s;
 	int i, len;
 
 	for (i = len = 0; s[i]; i++, len++) {
+		if (iskanji1(s, i)) {
+			i++;
+			len++;
+		}
 #ifdef	CODEEUC
-		if (isekana(s, i)) i++;
-		else
+		else if (isekana(s, i)) i++;
+#else
+		else if (isskana(s, i)) /*EMPTY*/;
 #endif
-		if (isctl(s[i])) len++;
+		else if (iscntrl2(s[i])) len++;
+		else if (ismsb(s[i])) len += 3;
 	}
+
 	return(len);
 }
 
 int atoi2(s)
 char *s;
 {
-	long n;
+	int n;
 
-	if (!s || !(s = evalnumeric(s, &n, 0)) || *s) return(-1);
-#if	MSDOS
-	if (n > MAXTYPE(int)) return(-1);
-#endif
-	return((int)n);
+	if (!sscanf2(s, "%<d%$", &n)) return(-1);
+
+	return(n);
 }
 
 #ifdef	USESTDARGH
-/*VARARGS2*/
-int fprintf2(FILE *fp, CONST char *fmt, ...)
+/*VARARGS1*/
+char *asprintf3(CONST char *fmt, ...)
 #else
-/*VARARGS2*/
-int fprintf2(fp, fmt, va_alist)
-FILE *fp;
+/*VARARGS1*/
+char *asprintf3(fmt, va_alist)
 CONST char *fmt;
 va_dcl
 #endif
 {
 	va_list args;
-	char *buf;
+	char *cp;
 	int n;
 
-#ifdef	USESTDARGH
-	va_start(args, fmt);
-#else
-	va_start(args);
-#endif
-
-	n = vasprintf2(&buf, fmt, args);
+	VA_START(args, fmt);
+	n = vasprintf2(&cp, fmt, args);
 	va_end(args);
 	if (n < 0) error("malloc()");
-	fputs(buf, fp);
-	free(buf);
-	return(n);
-}
 
-#ifdef	USESTDARGH
-/*VARARGS3*/
-int snprintf2(char *s, int size, CONST char *fmt, ...)
-#else
-/*VARARGS3*/
-int snprintf2(s, size, fmt, va_alist)
-char *s;
-int size;
-CONST char *fmt;
-va_dcl
-#endif
-{
-	va_list args;
-	char *buf;
-	int n;
-
-#ifdef	USESTDARGH
-	va_start(args, fmt);
-#else
-	va_start(args);
-#endif
-
-	n = vasprintf2(&buf, fmt, args);
-	va_end(args);
-	if (n < 0) error("malloc()");
-	strncpy2(s, buf, size - 1);
-	free(buf);
-	return(n);
+	return(cp);
 }
 
 VOID perror2(s)
@@ -733,14 +709,10 @@ char *s;
 	int duperrno;
 
 	duperrno = errno;
-	if (s) {
-		kanjifputs(s, stderr);
-		fputs(": ", stderr);
-	}
+	if (s) fprintf2(stderr, "%k: ", s);
 	fputs(strerror2(duperrno), stderr);
 	if (isttyiomode) fputc('\r', stderr);
-	fputc('\n', stderr);
-	fflush(stderr);
+	fputnl(stderr);
 }
 
 #ifdef	_NOORIGSHELL
@@ -756,6 +728,7 @@ char **envp;
 	for (i = 0; envp[i]; i++)
 		if (!strnenvcmp(envp[i], name, len) && envp[i][len] == '=')
 			break;
+
 	return(i);
 }
 
@@ -780,6 +753,7 @@ char *s, **envp;
 	new = (char **)realloc2(envp, (n + 2) * sizeof(char *));
 	new[n] = s;
 	new[n + 1] = (char *)NULL;
+
 	return(new);
 }
 #endif	/* _NOORIGSHELL */
@@ -811,6 +785,7 @@ char *name;
 	&& (cp = getshellvar(&(name[FDESIZ]), -1)))
 		return(cp);
 #endif	/* !_NOORIGSHELL */
+
 	return(NULL);
 }
 
@@ -844,12 +819,16 @@ int export;
 #ifdef	_NOORIGSHELL
 	if (export) environ = _putenv2(cp, environ);
 	else environ2 = _putenv2(cp, environ2);
+# ifndef	_NOPTY
+	sendparent(TE_PUTSHELLVAR, name, value, export);
+# endif
 #else	/* !_NOORIGSHELL */
 	if (((export) ? putexportvar(cp, len) : putshellvar(cp, len)) < 0) {
 		free(cp);
 		return(-1);
 	}
 #endif	/* !_NOORIGSHELL */
+
 	return(0);
 }
 
@@ -865,58 +844,63 @@ sigcst_t func;
 	sigemptyset(&(act.sa_mask));
 	sigemptyset(&(oact.sa_mask));
 	if (sigaction(sig, &act, &oact) < 0) return(SIG_ERR);
+
 	return(oact.sa_handler);
 }
 #endif	/* USESIGACTION */
 
-int system2(command, noconf)
+int system2(command, flags)
 char *command;
-int noconf;
+int flags;
 {
-	int n, wastty, mode, nl, ret;
+	int n, wastty, mode, ret;
 
 	if (!command || !*command) return(0);
 #ifdef	FAKEUNINIT
-	mode = nl = 0;		/* fake for -Wuninitialized */
+	mode = 0;		/* fake for -Wuninitialized */
 #endif
 	n = sigvecset(0);
 	if ((wastty = isttyiomode)) {
-		if (noconf >= 0) {
-			locate(0, n_line - 1);
-			putterm(l_clear);
+		if (!(flags & F_ISARCH)) {
+			Xlocate(0, n_line - 1);
+			Xputterm(L_CLEAR);
 		}
-		if (n && noconf > 0) mode = termmode(0);
-		nl = stdiomode();
+		if (n && (flags & F_NOCONFIRM)) mode = Xtermmode(0);
+		Xstdiomode();
 	}
 
 	ret = dosystem(command);
+	LOG1(_LOG_NOTICE_, ret, "system(\"%s\");", command);
+#ifdef	_NOORIGSHELL
+	checkscreen(-1, -1);
+#endif
 	sigvecset(n);
-	if (ret >= 127 && noconf > 0) {
+	if (ret >= 127 && (flags & F_NOCONFIRM)) {
 		if (dumbterm <= 2) fputc('\007', stderr);
-		fputc('\n', stderr);
-		kanjifputs(HITKY_K, stderr);
+		fprintf2(stderr, "\n%k", HITKY_K);
 		fflush(stderr);
-		ttyiomode(1);
+		Xttyiomode(1);
 		keyflush();
 		getkey2(0);
-		stdiomode();
-		fputc('\n', stderr);
+		Xstdiomode();
+		fputnl(stderr);
 	}
 
 	if (wastty) {
-		ttyiomode(nl);
-		if (n && noconf > 0) termmode(mode);
-		if (!noconf || (noconf < 0 && ret >= 127)) {
+		Xttyiomode(wastty - 1);
+		if (n && (flags & F_NOCONFIRM)) Xtermmode(mode);
+		if (!(flags & (F_NOCONFIRM | F_ISARCH))
+		|| ((flags & F_ISARCH) && ret >= 127)) {
 			hideclock = 1;
 			warning(0, HITKY_K);
 		}
 	}
+
 	return(ret);
 }
 
-/*ARGSUSED*/
-FILE *popen2(command, type)
-char *command, *type;
+FILE *popen2(command)
+char *command;
 {
 	FILE *fp;
 	int n;
@@ -925,45 +909,33 @@ char *command, *type;
 	n = sigvecset(0);
 	wasttyflags = 0;
 	if (isttyiomode) {
-		wasttyflags |= TF_TTYIOMODE;
-		if (stdiomode()) wasttyflags |= TF_TTYNL;
+		wasttyflags |= F_TTYIOMODE;
+		if (isttyiomode > 1) wasttyflags |= F_TTYNL;
+		Xstdiomode();
 	}
 
-#ifdef	_NOORIGSHELL
-# ifdef	DEBUG
-	_mtrace_file = "popen(start)";
-	fp = Xpopen(command, type);
-	if (_mtrace_file) _mtrace_file = NULL;
-	else {
-		_mtrace_file = "popen(end)";
-		malloc(0);	/* dummy malloc */
-	}
-# else
-	fp = Xpopen(command, type);
-# endif
-#else	/* !_NOORIGSHELL */
 	fp = dopopen(command);
-#endif	/* !_NOORIGSHELL */
 	sigvecset(n);
 	if (fp) {
-		if (wasttyflags & TF_TTYIOMODE) {
-			putterms(t_keypad);
-			tflush();
+		if (wasttyflags & F_TTYIOMODE) {
+			Xputterm(T_KEYPAD);
+			Xtflush();
 		}
 	}
 	else {
 		if (dumbterm <= 2) fputc('\007', stderr);
-		fputc('\n', stderr);
+		fputnl(stderr);
 		perror2(command);
 		fflush(stderr);
-		ttyiomode(1);
+		Xttyiomode(1);
 		keyflush();
 		getkey2(0);
-		stdiomode();
-		fputc('\n', stderr);
-		if (wasttyflags & TF_TTYIOMODE)
-			ttyiomode((wasttyflags & TF_TTYNL) ? 1 : 0);
+		Xstdiomode();
+		fputnl(stderr);
+		if (wasttyflags & F_TTYIOMODE)
+			Xttyiomode((wasttyflags & F_TTYNL) ? 1 : 0);
 	}
+
 	return(fp);
 }
 
@@ -972,13 +944,10 @@ FILE *fp;
 {
 	int ret;
 
-#ifdef	_NOORIGSHELL
-	ret = Xpclose(fp);
-#else
 	ret = dopclose(fp);
-#endif
-	if (wasttyflags & TF_TTYIOMODE)
-		ttyiomode((wasttyflags & TF_TTYNL) ? 1 : 0);
+	if (wasttyflags & F_TTYIOMODE)
+		Xttyiomode((wasttyflags & F_TTYNL) ? 1 : 0);
+
 	return(ret);
 }
 
@@ -987,11 +956,21 @@ char *getwd2(VOID_A)
 	char cwd[MAXPATHLEN];
 
 	if (!Xgetwd(cwd)) lostcwd(cwd);
+
 	return(strdup2(cwd));
 }
 
 #if	!MSDOS && !defined (NOTZFILEH) \
 && !defined (USEMKTIME) && !defined (USETIMELOCAL)
+static long NEAR char2long(s)
+u_char *s;
+{
+	return((long)((u_long)(s[3])
+		| ((u_long)(s[2]) << (BITSPERBYTE * 1))
+		| ((u_long)(s[1]) << (BITSPERBYTE * 2))
+		| ((u_long)(s[0]) << (BITSPERBYTE * 3))));
+}
+
 static int NEAR tmcmp(tm1, tm2)
 struct tm *tm1, *tm2;
 {
@@ -1005,11 +984,46 @@ struct tm *tm1, *tm2;
 		return (tm1 -> tm_hour - tm2 -> tm_hour);
 	if (tm1 -> tm_min != tm2 -> tm_min)
 		return (tm1 -> tm_min - tm2 -> tm_min);
+
 	return (tm1 -> tm_sec - tm2 -> tm_sec);
 }
 #endif	/* !MSDOS && !NOTZFILEH && !USEMKTIME && !USETIMELOCAL */
 
 #if	!defined (USEMKTIME) && !defined (USETIMELOCAL)
+static time_t NEAR timegm2(tm)
+struct tm *tm;
+{
+	time_t t;
+	int i, y;
+
+	y = (tm -> tm_year < 1900) ? tm -> tm_year + 1900 : tm -> tm_year;
+
+	t = ((long)y - 1970) * 365;
+	t += ((y - 1 - 1968) / 4)
+		- ((y - 1 - 1900) / 100)
+		+ ((y - 1 - 1600) / 400);
+	for (i = 1; i < tm -> tm_mon + 1; i++) switch (i) {
+		case 2:
+			if (!(y % 4) && ((y % 100) || !(y % 400))) t++;
+			t += 28;
+			break;
+		case 4:
+		case 6:
+		case 9:
+		case 11:
+			t += 30;
+			break;
+		default:
+			t += 31;
+			break;
+	}
+	t += tm -> tm_mday - 1;
+	t *= 60L * 60L * 24L;
+	t += ((long)(tm -> tm_hour) * 60L + tm -> tm_min) * 60L + tm -> tm_sec;
+
+	return(t);
+}
+
 static long NEAR gettimezone(tm, t)
 struct tm *tm;
 time_t t;
@@ -1020,25 +1034,21 @@ time_t t;
 	ftime(&buffer);
 	return((long)(buffer.timezone) * 60L);
 # else	/* !MSDOS */
-#  ifdef	NOTMGMTOFF
-	struct timeval t_val;
-	struct timezone t_zone;
-#  endif
 #  ifndef	NOTZFILEH
 	struct tzhead head;
-#  endif
-	struct tm tmbuf;
 	FILE *fp;
 	time_t tmp;
-	long i, tz, leap, nleap, ntime, ntype, nchar;
+	long i, leap, nleap, ntime, ntype, nchar;
 	char *cp, buf[MAXPATHLEN];
 	u_char c;
+#  endif
+	struct tm tmbuf;
+	long tz;
 
 	memcpy((char *)&tmbuf, (char *)tm, sizeof(struct tm));
 
 #  ifdef	NOTMGMTOFF
-	gettimeofday2(&t_val, &t_zone);
-	tz = t_zone.tz_minuteswest * 60L;
+	tz = (long)t - (long)timegm2(localtime(&t));
 #  else
 	tz = -(localtime(&t) -> tm_gmtoff);
 #  endif
@@ -1048,7 +1058,7 @@ time_t t;
 	if (!cp || !*cp) cp = TZDEFAULT;
 	if (cp[0] == _SC_) strcpy(buf, cp);
 	else strcatdelim2(buf, TZDIR, cp);
-	if (!(fp = fopen(buf, "r"))) return(tz);
+	if (!(fp = fopen(buf, "rb"))) return(tz);
 	if (fread(&head, sizeof(struct tzhead), 1, fp) != 1) {
 		fclose(fp);
 		return(tz);
@@ -1072,8 +1082,8 @@ time_t t;
 	}
 	if (i > 0) {
 		i--;
-		i *= sizeof(char);
-		i += sizeof(struct tzhead) + ntime * sizeof(char) * 4;
+		i *= (int)sizeof(char);
+		i += (int)sizeof(struct tzhead) + ntime * 4 * sizeof(char);
 		if (fseek(fp, i, 0) < 0
 		|| fread(&c, sizeof(char), 1, fp) != 1) {
 			fclose(fp);
@@ -1081,8 +1091,8 @@ time_t t;
 		}
 		i = c;
 	}
-	i *= sizeof(char) * (4 + 1 + 1);
-	i += sizeof(struct tzhead) + ntime * sizeof(char) * (4 + 1);
+	i *= (4 + 1 + 1) * sizeof(char);
+	i += (int)sizeof(struct tzhead) + ntime * (4 + 1) * sizeof(char);
 	if (fseek(fp, i, 0) < 0
 	|| fread(buf, sizeof(char), 4, fp) != 4) {
 		fclose(fp);
@@ -1091,8 +1101,8 @@ time_t t;
 	tmp = char2long(buf);
 	tz = -tmp;
 
-	i = sizeof(struct tzhead) + ntime * sizeof(char) * (4 + 1)
-		+ ntype * sizeof(char) * (4 + 1 + 1)
+	i = (int)sizeof(struct tzhead) + ntime * (4 + 1) * sizeof(char)
+		+ ntype * (4 + 1 + 1) * sizeof(char)
 		+ nchar * sizeof(char);
 	if (fseek(fp, i, 0) < 0) {
 		fclose(fp);
@@ -1128,12 +1138,14 @@ time_t time2(VOID_A)
 	struct timeb buffer;
 
 	ftime(&buffer);
+
 	return((time_t)(buffer.time));
 #else
 	struct timeval t_val;
 	struct timezone tz;
 
 	gettimeofday2(&t_val, &tz);
+
 	return((time_t)(t_val.tv_sec));
 #endif
 }
@@ -1148,37 +1160,9 @@ struct tm *tm;
 # ifdef	USETIMELOCAL
 	return(timelocal(tm));
 # else	/* !USETIMELOCAL */
-	time_t d, t;
-	int i, y;
+	time_t t;
 
-	y = (tm -> tm_year < 1900) ? tm -> tm_year + 1900 : tm -> tm_year;
-
-	d = ((long)y - 1970) * 365;
-	d += ((y - 1 - 1968) / 4)
-		- ((y - 1 - 1900) / 100)
-		+ ((y - 1 - 1600) / 400);
-	for (i = 1; i < tm -> tm_mon + 1; i++) {
-		switch (i) {
-			case 2:
-				if (!(y % 4)
-				&& ((y % 100)
-				|| !(y % 400))) d++;
-				d += 28;
-				break;
-			case 4:
-			case 6:
-			case 9:
-			case 11:
-				d += 30;
-				break;
-			default:
-				d += 31;
-				break;
-		}
-	}
-	d += tm -> tm_mday - 1;
-	t = ((long)(tm -> tm_hour) * 60L + tm -> tm_min) * 60L + tm -> tm_sec;
-	t += d * 60L * 60L * 24L;
+	t = timegm2(tm);
 	t += gettimezone(tm, t);
 
 	return(t);
@@ -1211,5 +1195,6 @@ int nulcnv;
 	if (i > 0 && cp[i - 1] == '\r') i--;
 #endif
 	cp[i++] = '\0';
+
 	return(realloc2(cp, i));
 }
