@@ -73,8 +73,9 @@ extern char *unitblpath;
 #endif
 #ifndef	_NOPTY
 extern int ptymode;
+extern int ptyinternal;
 extern int parentfd;
-#endif	/* !_NOPTY */
+#endif
 
 #define	CLOCKUPDATE	10	/* sec */
 
@@ -148,6 +149,10 @@ static VOID NEAR printtime __P_((int));
 static int trapalrm __P_((VOID_A));
 #endif
 static char *NEAR getversion __P_((int *));
+#ifndef	_NOLOGGING
+static VOID NEAR startlog __P_((char **));
+static VOID NEAR endlog __P_((int));
+#endif
 #ifdef	_NOORIGSHELL
 static int NEAR execruncomline __P_((char *, char *, int, char *));
 #endif
@@ -165,6 +170,9 @@ char *tmpfilename = NULL;
 int adjtty = 0;
 #endif
 int showsecond = 0;
+#if	FD >= 2
+int thruargs = 0;
+#endif
 int hideclock = 0;
 #ifdef	SIGALRM
 int noalrm = 1;
@@ -267,6 +275,9 @@ int sig;
 		forcecleandir(deftmpdir, tmpfilename);
 #ifndef	_NODOSDRIVE
 		dosallclose();
+#endif
+#ifndef	_NOLOGGING
+		endlog(sig + 128);
 #endif
 		endterm();
 		inittty(1);
@@ -660,7 +671,8 @@ int hide;
 	}
 	if (timersec-- < CLOCKUPDATE && !showsecond) return;
 #ifndef	_NOPTY
-	if (isptymode() && hide <= 1) hide = 0;
+	if (checkallpty() < 0) rewritefile(0);
+	if (isptymode() && !ptyinternal && hide <= 1) hide = 0;
 #endif
 	if (hide) return;
 
@@ -696,6 +708,12 @@ static int trapalrm(VOID_A)
 
 	duperrno = errno;
 	signal2(SIGALRM, SIG_IGN);
+# if	!defined (_NOORIGSHELL) && !defined (_NOPTY)
+	if (isshptymode()) {
+		if (mypid == shellpid) printtime(hideclock);
+	}
+	else
+# endif
 	if (isorgpid()) printtime(hideclock);
 # ifdef	SIGWINCH
 	pollscreen(0);
@@ -789,12 +807,12 @@ VOID title(VOID_A)
 	Xcputs2(" Ver.");
 	len += 5;
 	cp = getversion(&i);
-	Xcprintf2("%-*.*s", i, i, cp);
+	cputstr(i, cp);
 	if (distributor) {
 		Xputch2('#');
 		i++;
 	}
-	cp = (iswellomit()) ? "" : " (c)1995-2005 T.Shirai  ";
+	cp = (iswellomit()) ? nullstr : " (c)1995-2006 T.Shirai  ";
 	Xcputs2(cp);
 	i = n_column - len - strlen2(cp) - i;
 	while (i-- > 0) Xputch2(' ');
@@ -823,6 +841,61 @@ VOID saveorigenviron(VOID_A)
 }
 #endif	/* !_NOCUSTOMIZE */
 
+#ifndef	_NOLOGGING
+static VOID NEAR startlog(argv)
+char **argv;
+{
+# ifndef	NOUID
+	uid_t uid;
+	int lvl;
+# endif
+	char *cp, *tmp, buf[MAXLOGLEN + 1];
+	int i, len;
+
+	cp = buf;
+	for (i = 1; argv && argv[i]; i++) {
+		if (cp > buf) *(cp++) = ' ';
+		tmp = killmeta(argv[i]);
+		len = snprintf2(cp, (int)sizeof(buf) - (cp - buf), "%s", tmp);
+		free(tmp);
+		if (len < 0) break;
+		cp += len;
+	}
+	*cp = '\0';
+
+	cp = getversion(&len);
+# ifdef	NOUID
+	logmessage(_LOG_DEBUG_, "%s (%-*.*s) starts; PWD=%k; ARGS=%k",
+		progname, len, len, cp, origpath, buf);
+# else
+	lvl = ((uid = getuid())) ? _LOG_DEBUG_ : _LOG_WARNING_;
+	logmessage(lvl, "%s (%-*.*s) starts; UID=%d; PWD=%k; ARGS=%k",
+		progname, len, len, cp, uid, origpath, buf);
+# endif
+}
+
+static VOID NEAR endlog(status)
+int status;
+{
+# ifndef	NOUID
+	uid_t uid;
+	int lvl;
+# endif
+	char cwd[MAXPATHLEN];
+
+	if (!Xgetwd(cwd)) strcpy(cwd, "?");
+# ifdef	NOUID
+	logmessage(_LOG_DEBUG_, "%s ends; PWD=%k; STATUS=%d",
+		progname, cwd, status);
+# else
+	lvl = ((uid = getuid())) ? _LOG_DEBUG_ : _LOG_WARNING_;
+	logmessage(lvl, "%s ends; UID=%d; PWD=%k; STATUS=%d",
+		progname, uid, cwd, status);
+# endif
+	logclose();
+}
+#endif	/* !_NOLOGGING */
+
 #ifdef	_NOORIGSHELL
 static int NEAR execruncomline(command, file, n, line)
 char *command, *file;
@@ -1077,6 +1150,9 @@ char **argv;
 		loadhistory(0, histfile);
 		entryhist(1, origpath, 1);
 	}
+#ifndef	_NOLOGGING
+	startlog(argv);
+#endif
 }
 
 /*ARGSUSED*/
@@ -1085,13 +1161,16 @@ int status;
 {
 	char cwd[MAXPATHLEN];
 
+#ifndef	_NOLOGGING
+	endlog(status);
+#endif
 	if (interactive && savehist > 0) savehistory(0, histfile);
 
 	cwd[0] = '\0';
 	if (origpath && _chdir2(origpath) < 0) {
 		perror2(origpath);
 		if (!Xgetwd(cwd)) *cwd = '\0';
-		rawchdir(_SS_);
+		rawchdir(rootpath);
 	}
 	free(origpath);
 	free(progname);
@@ -1336,8 +1415,15 @@ char *argv[], *envp[];
 #ifdef	SIGWINCH
 	nowinch = 0;
 #endif
+#if	FD >= 2
+	main_fd(&(argv[i]), (thruargs) ? 1 : 0);
+#else
 	main_fd(&(argv[i]), 0);
+#endif
 	sigvecset(0);
+#ifndef	_NOLOGGING
+	endlog(0);
+#endif
 
 	Xstdiomode();
 #if	!defined (_NOORIGSHELL) && !defined (NOJOB)
