@@ -55,10 +55,6 @@ typedef struct _kstree_t {
 } kstree_t;
 #endif	/* !MSDOS */
 
-#ifdef	USESELECTH
-#include <sys/select.h>
-#endif
-
 #ifndef	NOUNISTDH
 #include <unistd.h>
 #endif
@@ -77,7 +73,7 @@ extern int errno;
 #include "termio.h"
 #include "term.h"
 
-#define	WINTERMNAME	"iris"
+#define	WINTERMNAME		"iris"
 #ifdef	USESTRERROR
 #define	strerror2		strerror
 #else
@@ -338,16 +334,6 @@ extern int tputs __P_((char *, int, int (*)__P_((tputs_t))));
 #endif
 #endif	/* !MSDOS */
 
-#ifndef	FD_SET
-typedef struct fd_set {
-	u_int fds_bits[1];
-} fd_set;
-# define	FD_ZERO(p)	(((p) -> fds_bits[0]) = 0)
-# define	FD_SET(n, p)	(((p) -> fds_bits[0]) |= ((u_int)1 << (n)))
-#endif	/* !FD_SET */
-
-#define	MAXFDSET	256
-
 static int NEAR err2 __P_((char *));
 #if	!MSDOS
 static char *NEAR tstrdup __P_((char *));
@@ -411,6 +397,7 @@ int (*keywaitfunc)__P_((VOID_A)) = NULL;
 #if	!MSDOS
 int usegetcursor = 0;
 int suspended = 0;
+char *duptty[2] = {NULL, NULL};
 #endif
 int ttyio = -1;
 int isttyiomode = 0;
@@ -459,10 +446,10 @@ static int specialkeycode[] = {
 	K_F(11), K_F(12), K_F(13), K_F(14), K_F(15),
 	K_F(16), K_F(17), K_F(18), K_F(19), K_F(20), K_HELP
 };
-#define	SPECIALKEYSIZ	((int)(sizeof(specialkeycode) / sizeof(int)))
+#define	SPECIALKEYSIZ	arraysize(specialkeycode)
 #else	/* !MSDOS */
 static char *dumblist[] = {"dumb", "un", "unknown"};
-#define	DUMBLISTSIZE	((int)sizeof(dumblist) / sizeof(char *))
+#define	DUMBLISTSIZE	arraysize(dumblist)
 static keyseq_t keyseq[K_MAX - K_MIN + 1];
 static kstree_t *keyseqtree = NULL;
 static char *defkeyseq[K_MAX - K_MIN + 1] = {
@@ -1077,12 +1064,34 @@ int keyflush(VOID_A)
 
 	return(0);
 }
+
+int savettyio(reset)
+int reset;
+{
+	char *tty;
+	int n;
+
+	if (reset) tty = NULL;
+	else {
+		savetermio(ttyio, &tty, NULL);
+		if (!tty) return(-1);
+	}
+
+	n = (isttyiomode) ? 0 : 1;
+	if (duptty[n]) free(duptty[n]);
+	duptty[n] = tty;
+
+	return(n);
+}
 #endif	/* !MSDOS */
 
 int ttyiomode(isnl)
 int isnl;
 {
 	if (ttyio < 0) /*EMPTY*/;
+#if	!MSDOS
+	else if (duptty[0]) loadtermio(ttyio, duptty[0], NULL);
+#endif
 	else {
 #if	MSDOS
 		raw2();
@@ -1092,11 +1101,14 @@ int isnl;
 		noecho2();
 		nonl2();
 		notabs();
+		if (dumbterm > 1) ttymode(ECHO | CRMOD, 0, 0, 0);
 # else	/* !USESGTTY */
 		if (isnl) ttymode(0, TIO_LCOOKED | ~TIO_LNOECHO,
 			IGNBRK, TIO_ICOOKED | ICRNL, TIO_ONL | TAB3, 0, 1, 0);
 		else ttymode(0, TIO_LCOOKED | ~TIO_LNOECHO,
 			IGNBRK, TIO_ICOOKED | ICRNL, TAB3, TIO_ONL, 1, 0);
+		if (dumbterm > 1)
+			ttymode(ECHO, 0, ICRNL, 0, TIO_ONL, 0, 1, 0);
 # endif	/* !USESGTTY */
 #endif	/* !MSDOS */
 	}
@@ -1115,6 +1127,9 @@ int stdiomode(VOID_A)
 
 	isnl = (isttyiomode) ? isttyiomode - 1 : 0;
 	if (ttyio < 0) /*EMPTY*/;
+#if	!MSDOS
+	else if (duptty[1]) loadtermio(ttyio, duptty[1], NULL);
+#endif
 	else {
 #if	MSDOS
 		cooked2();
@@ -1339,7 +1354,7 @@ int *xp, *yp;
 	format = SIZEFMT;
 	keyflush();
 # if	MSDOS
-	for (i = 0; i < (int)sizeof(GETSIZE) - 1; i++)
+	for (i = 0; i < strsize(GETSIZE); i++)
 		bdos(0x06, GETSIZE[i], 0);
 # else
 	if (!usegetcursor) return(-1);
@@ -1359,7 +1374,7 @@ int *xp, *yp;
 # endif
 	} while (buf[i] != format[0]);
 
-	if (buf[i] == format[0]) while (i < (int)sizeof(buf) - 2) {
+	if (buf[i] == format[0]) while (i < strsize(buf) - 1) {
 		if (!kbhit2(WAITKEYPAD * 1000L)) break;
 # if	MSDOS
 		buf[++i] = bdos(0x07, 0x00, 0);
@@ -1367,10 +1382,10 @@ int *xp, *yp;
 		if ((tmp = getch2()) == EOF) break;
 		buf[++i] = tmp;
 # endif
-		if (buf[i] == format[sizeof(SIZEFMT) - 2]) break;
+		if (buf[i] == format[strsize(SIZEFMT) - 1]) break;
 	}
 	keyflush();
-	if (!i || buf[i] != format[sizeof(SIZEFMT) - 2]) return(-1);
+	if (!i || buf[i] != format[strsize(SIZEFMT) - 1]) return(-1);
 	buf[++i] = '\0';
 
 	count = 0;
@@ -1551,8 +1566,18 @@ char *s;
 	char *p;
 
 	p = strbuf;
+#  ifdef	DEBUG
+	_mtrace_file = "tgetstr(start)";
 	s = tgetstr(s, &p);
-# endif
+	if (_mtrace_file) _mtrace_file = NULL;
+	else {
+		_mtrace_file = "tgetstr(end)";
+		malloc(0);	/* dummy alloc */
+	}
+#  else
+	s = tgetstr(s, &p);
+#  endif
+# endif	/* !USETERMINFO */
 	if (s) {
 		if (*term) free(*term);
 		*term = tstrdup(s);
@@ -1567,14 +1592,35 @@ char **term, *str1, *str2;
 # ifdef	USETERMINFO
 	if (str1) str1 = tstrdup(str1);
 	else str1 = tparamstr(str2, 1, 1);
-# else
+# else	/* !USETERMINFO */
 	char *p, strbuf[TERMCAPSIZE];
 
 	p = strbuf;
 
+#  ifdef	DEBUG
+	_mtrace_file = "tgetstr(start)";
+	str1 = tgetstr(str1, &p);
+	if (_mtrace_file) _mtrace_file = NULL;
+	else {
+		_mtrace_file = "tgetstr(end)";
+		malloc(0);	/* dummy alloc */
+	}
+	if (str1) str1 = tstrdup(str1);
+	else {
+		_mtrace_file = "tgetstr(start)";
+		str2 = tgetstr(str2, &p);
+		if (_mtrace_file) _mtrace_file = NULL;
+		else {
+			_mtrace_file = "tgetstr(end)";
+			malloc(0);	/* dummy alloc */
+		}
+		str1 = tparamstr(str2, 1, 1);
+	}
+#  else
 	if ((str1 = tgetstr(str1, &p))) str1 = tstrdup(str1);
 	else str1 = tparamstr(tgetstr(str2, &p), 1, 1);
-# endif
+#  endif
+# endif	/* !USETERMINFO */
 	if (str1) {
 		if (*term) free(*term);
 		*term = str1;
@@ -1620,8 +1666,8 @@ int num;
 	int i, n;
 
 	if (!parent || !(parent -> next))
-		new = (kstree_t *)malloc(sizeof(kstree_t) * num);
-	else new = (kstree_t *)realloc(parent -> next, sizeof(kstree_t) * num);
+		new = (kstree_t *)malloc(num * sizeof(kstree_t));
+	else new = (kstree_t *)realloc(parent -> next, num * sizeof(kstree_t));
 	if (!new) err2("realloc()");
 
 	if (!parent) n = 0;
@@ -1728,7 +1774,7 @@ char *s;
 		if (i < DUMBLISTSIZE) dumbterm = 1;
 	}
 # ifdef	IRIX
-	winterm = !strncmp(term, WINTERMNAME, sizeof(WINTERMNAME) - 1);
+	winterm = !strncmp(term, WINTERMNAME, strsize(WINTERMNAME));
 # endif
 
 	for (;;) {
@@ -2571,13 +2617,12 @@ long usec;
 {
 #if	defined (NOTUSEBIOS) || !defined (DJGPP) || (DJGPP < 2)
 	union REGS reg;
-#else
+#else	/* !NOTUSEBIOS && DJGPP && DJGPP >= 2 */
 # ifndef	PC98
-	fd_set readfds;
 	struct timeval tv;
-	int n, fd;
+	int n;
 # endif
-#endif
+#endif	/* !NOTUSEBIOS && DJGPP && DJGPP >= 2 */
 
 	if (ungetnum > 0) return(1);
 #ifdef	NOTUSEBIOS
@@ -2598,17 +2643,9 @@ long usec;
 	return(reg.h.bh != 0);
 # else	/* !PC98 */
 #  if	defined (DJGPP) && (DJGPP >= 2)
-	fd = (ttyio < MAXFDSET || (n = safe_dup(ttyio)) < 0) ? ttyio : n;
 	tv.tv_sec = (time_t)usec / (time_t)1000000;
 	tv.tv_usec = (time_t)usec % (time_t)1000000;
-	FD_ZERO(&readfds);
-	FD_SET(fd, &readfds);
-
-	do {
-		n = select(fd + 1, &readfds, NULL, NULL, &tv);
-	} while (n < 0 && errno == EINTR);
-	if (fd != ttyio) close(fd);
-	if (n < 0) err2("select()");
+	if ((n = readselect(1, &ttyio, NULL, &tv)) < 0) err2("select()");
 
 	return(n);
 #  else	/* !DJGPP || DJGPP < 2 */
@@ -2731,7 +2768,7 @@ int sig;
 			raise(sig);
 		}
 # endif
-		if (keywaitfunc && (*keywaitfunc)() < 0) return(-1);
+		if (keywaitfunc && (ch = (*keywaitfunc)()) < 0) return(ch);
 	} while (!i);
 
 #else	/* !DJGPP || NOTUSEBIOS || PC98 */
@@ -2744,7 +2781,7 @@ int sig;
 # endif
 			memcpy((char *)tbuf1, (char *)tbuf2, sizeof(tbuf1));
 		}
-		if (keywaitfunc && (*keywaitfunc)() < 0) return(-1);
+		if (keywaitfunc && (ch = (*keywaitfunc)()) < 0) return(ch);
 	}
 #endif	/* !DJGPP || NOTUSEBIOS || PC98 */
 	if ((ch = getch2()) == EOF) return(K_NOKEY);
@@ -2778,9 +2815,12 @@ int sig;
 int ungetch2(c)
 int c;
 {
-	if (ungetnum >= (int)sizeof(ungetbuf) / sizeof(u_char) - 1)
-		return(EOF);
-	ungetbuf[ungetnum++] = c;
+	if (ungetnum >= arraysize(ungetbuf) - 1) return(EOF);
+	if (ungetnum)
+		memmove(&(ungetbuf[1]), &(ungetbuf[0]),
+			ungetnum * sizeof(u_char));
+	ungetbuf[0] = c;
+	ungetnum++;
 
 	return(c);
 }
@@ -2878,21 +2918,12 @@ long usec;
 # ifdef	NOSELECT
 	return((usec) ? 1 : 0);
 # else
-	fd_set readfds;
 	struct timeval tv;
-	int n, fd;
+	int n;
 
-	fd = (ttyio < MAXFDSET || (n = safe_dup(ttyio)) < 0) ? ttyio : n;
 	tv.tv_sec = (time_t)usec / (time_t)1000000;
 	tv.tv_usec = (time_t)usec % (time_t)1000000;
-	FD_ZERO(&readfds);
-	FD_SET(fd, &readfds);
-
-	do {
-		n = select(fd + 1, &readfds, NULL, NULL, &tv);
-	} while (n < 0 && errno == EINTR);
-	if (fd != ttyio) close(fd);
-	if (n < 0) err2("select()");
+	if ((n = readselect(1, &ttyio, NULL, &tv)) < 0) err2("select()");
 
 	return(n);
 # endif
@@ -2931,7 +2962,7 @@ int sig;
 			ttyiomode((isttyiomode) ? isttyiomode - 1 : 0);
 			suspended = 0;
 		}
-		if (keywaitfunc && (*keywaitfunc)() < 0) return(-1);
+		if (keywaitfunc && (ch = (*keywaitfunc)()) < 0) return(ch);
 # ifndef	TIOCSTI
 		if (ungetnum > 0) return((int)ungetbuf[--ungetnum]);
 # endif
@@ -2962,10 +2993,12 @@ int sig;
 			return(mkmetakey(ch));
 		for (j = 0; j < p -> num; j++)
 			if (key == keyseq[p -> next[j].key].str[0]) break;
-		if (j >= p -> num) return(key);
+		if (j >= p -> num) {
+			ungetch2(ch);
+			return(key);
+		}
 		p = &(p -> next[j]);
-		if (keyseq[p -> key].len == 1)
-			return(keyseq[p -> key].code);
+		if (keyseq[p -> key].len == 1) return(keyseq[p -> key].code);
 	}
 	else {
 		for (j = 0; j < p -> num; j++)
@@ -2981,14 +3014,19 @@ int sig;
 	for (i = 1; p && p -> next; i++) {
 		for (j = 0; j < p -> num; j++)
 			if (ch == keyseq[p -> next[j].key].str[i]) break;
-		if (j >= p -> num) return(key);
+		if (j >= p -> num) break;
 		p = &(p -> next[j]);
 		if (keyseq[p -> key].len == i + 1)
 			return(keyseq[p -> key].code);
 		if (!kbhit2(WAITKEYPAD * 1000L) || (ch = getch2()) == EOF)
-			return(key);
+			break;
 	}
 
+	for (j = 1; j < i; j++) {
+		if (j >= keyseq[p -> key].len) break;
+		ungetch2(keyseq[p -> key].str[j]);
+	}
+	ungetch2(ch);
 	return(key);
 }
 
@@ -3015,9 +3053,12 @@ int c;
 	ch = c;
 	Xioctl(ttyio, TIOCSTI, &ch);
 # else
-	if (ungetnum >= (int)sizeof(ungetbuf) / sizeof(u_char) - 1)
-		return(EOF);
-	ungetbuf[ungetnum++] = c;
+	if (ungetnum >= arraysize(ungetbuf) - 1) return(EOF);
+	if (ungetnum)
+		memmove(&(ungetbuf[1]), &(ungetbuf[0]),
+			ungetnum * sizeof(u_char));
+	ungetbuf[0] = c;
+	ungetnum++;
 # endif
 
 	return(c);
@@ -3093,7 +3134,8 @@ int xmax, ymax;
 
 	if (dumbterm) /*EMPTY*/;
 	else if (usegetcursor || x < 0 || y < 0) {
-		VOID_C setscroll(-1, -1);
+		if (usegetcursor) VOID_C setscroll(-1, -1);
+		else VOID_C setscroll(0, 998);
 		if (maxlocate(&ty, &tx) >= 0 && (tx > x || ty > y)) {
 			if (tx > x) x = tx;
 			if (ty > y) y = ty;
@@ -3112,7 +3154,7 @@ int xmax, ymax;
 	if (n_line <= 0 || (ymax > 0 && n_line < ymax))
 		return("Line size too small");
 
-	if (xmax > 0 && ymax > 0) VOID_C setscroll(-1, -1);
+	if (xmax > 0 && ymax > 0) VOID_C setscroll(0, n_line - 1);
 
 	return(NULL);
 }
