@@ -1,7 +1,7 @@
 /*
  *	unixdisk.c
  *
- *	UNIXlike Disk Access Module
+ *	UNIX-like disk accessing module
  */
 
 #include <stdio.h>
@@ -9,12 +9,8 @@
 #include <io.h>
 #include "machine.h"
 
-#ifdef	NOVOID
-#define	VOID
-#define	VOID_T	int
-#else
-#define	VOID	void
-#define	VOID_T	void
+#ifndef	NOUNISTDH
+#include <unistd.h>
 #endif
 
 #if	MSDOS && defined (_NOUSELFN) && !defined (_NODOSDRIVE)
@@ -23,7 +19,14 @@
 
 #include "unixdisk.h"
 #include "dosdisk.h"
+#include "kctype.h"
 
+#ifndef	ENODEV
+#define	ENODEV	EACCES
+#endif
+#ifndef	EIO
+#define	EIO	ENODEV
+#endif
 #ifndef	O_ACCMODE
 #define	O_ACCMODE	(O_RDONLY | O_WRONLY | O_RDWR)
 #endif
@@ -31,32 +34,7 @@
 #define	KC_SJIS1	0001
 #define	KC_SJIS2	0002
 #define	KC_EUCJP	0010
-
-#ifdef	LSI_C
-#include <jctype.h>
-#define	toupper2(c)	toupper(c)
-#define	tolower2(c)	tolower(c)
-#define	issjis1(c)	iskanji(c)
-#define	issjis2(c)	iskanji2(c)
-#else	/* !LSI_C */
-#define	toupper2(c)	(uppercase[(u_char)(c)])
-#define	tolower2(c)	(lowercase[(u_char)(c)])
-# ifndef	issjis1
-# define	issjis1(c)	(kctypetable[(u_char)(c)] & KC_SJIS1)
-# endif
-# ifndef	issjis2
-# define	issjis2(c)	(kctypetable[(u_char)(c)] & KC_SJIS2)
-# endif
-#endif	/* !LSI_C */
-
-#ifdef	CODEEUC
-# ifndef	iseuc
-# define	iseuc(c)	(kctypetable[(u_char)(c)] & KC_EUCJP)
-# endif
-#define	iskanji1(s, i)	(iseuc((s)[i]) && iseuc((s)[(i) + 1]))
-#else
-#define	iskanji1(s, i)	(issjis1((s)[i]) && issjis2((s)[(i) + 1]))
-#endif
+#define	int21call(reg, sreg)	intcall(0x21, reg, sreg)
 
 #ifdef	FD
 extern int _dospath __P_((char *));
@@ -66,6 +44,7 @@ extern int isdelim __P_((char *, int));
 extern char *strcatdelim __P_((char *));
 extern char *strcatdelim2 __P_((char *, char *, char *));
 extern int isdotdir __P_((char *));
+extern char curpath[];
 #else
 static int NEAR _dospath __P_((char *));
 static char *NEAR strdelim __P_((char *, int));
@@ -74,15 +53,13 @@ static int NEAR isdelim __P_((char *, int));
 static char *NEAR strcatdelim __P_((char *));
 static char *NEAR strcatdelim2 __P_((char *, char *, char *));
 static int NEAR isdotdir __P_((char *));
+static char curpath[] = ".";
 #endif
 
-static int NEAR seterrno __P_((u_short));
 #ifdef	DJGPP
 static int NEAR dos_putpath __P_((char *, int));
 #endif
 static char *NEAR duplpath __P_((char *));
-static long NEAR intcall __P_((int, __dpmi_regs *, struct SREGS *));
-#define	int21call(reg, sreg)	intcall(0x21, reg, sreg)
 #ifndef	_NODOSDRIVE
 static char *NEAR regpath __P_((char *, char *));
 static VOID NEAR biosreset __P_((int));
@@ -99,12 +76,13 @@ static int NEAR lockdrive __P_((int));
 static int NEAR unlockdrive __P_((int, int));
 static int NEAR dosrawdiskio __P_((int, u_long, u_char *, int, int, int));
 #endif
-static int NEAR dos_findfirst __P_((char *, u_short, struct dosfind_t *));
+static int NEAR dos_findfirst __P_((char *, u_int, struct dosfind_t *));
 static int NEAR dos_findnext __P_((struct dosfind_t *));
 #ifndef	_NOUSELFN
-static long NEAR lfn_findfirst __P_((char *, u_short, struct lfnfind_t *));
-static int NEAR lfn_findnext __P_((u_short, struct lfnfind_t *));
-static int NEAR lfn_findclose __P_((u_short));
+static int NEAR lfn_findfirst __P_((u_int *, char *,
+		u_int, struct lfnfind_t *));
+static int NEAR lfn_findnext __P_((u_int, struct lfnfind_t *));
+static int NEAR lfn_findclose __P_((u_int));
 #endif
 #ifndef	_NOUSELFN
 static int NEAR gendosname __P_((char *));
@@ -128,14 +106,6 @@ static int maxdrive = -1;
 static int dos7access = 0;
 #define	D7_DOSVER	017
 #define	D7_CAPITAL	020
-static u_short doserrlist[] = {
-	0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 15, 18, 65, 80
-};
-#define	DOSERRLISTSIZ	((int)(sizeof(doserrlist) / sizeof(u_short)))
-static int unixerrlist[] = {
-	0, EINVAL, ENOENT, ENOENT, EMFILE, EACCES,
-	EBADF, ENOMEM, ENOMEM, ENOMEM, ENODEV, 0, EACCES, EEXIST
-};
 #if	!defined (DJGPP) && !defined (_NODOSDRIVE)
 static u_char int25bin[] = {
 	0x1e,				/* 0000: push ds */
@@ -171,125 +141,14 @@ static u_char int25bin[] = {
 	0xcb				/* 002a: retf */
 };
 static int (far *doint25)__P_((VOID_A)) = NULL;
-#endif
-
-#ifndef	LSI_C
-# ifdef	FD
-extern u_char uppercase[256];
-extern u_char lowercase[256];
-extern u_char kctypetable[256];
-# else	/* !FD */
-static u_char uppercase[256] = {
-	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,	/* 0x00 */
-	0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
-	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,	/* 0x10 */
-	0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
-	0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,	/* 0x20 */
-	0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
-	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,	/* 0x30 */
-	0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
-	0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,	/* 0x40 */
-	0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
-	0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57,	/* 0x50 */
-	0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
-	0x60, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,	/* 0x60 */
-	0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
-	0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57,	/* 0x70 */
-	0x58, 0x59, 0x5a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f,
-	0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,	/* 0x80 */
-	0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
-	0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97,	/* 0x90 */
-	0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
-	0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,	/* 0xa0 */
-	0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
-	0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7,	/* 0xb0 */
-	0xb8, 0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf,
-	0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7,	/* 0xc0 */
-	0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
-	0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7,	/* 0xd0 */
-	0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
-	0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7,	/* 0xe0 */
-	0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
-	0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7,	/* 0xf0 */
-	0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff
-};
-static u_char lowercase[256] = {
-	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,	/* 0x00 */
-	0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
-	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,	/* 0x10 */
-	0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
-	0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,	/* 0x20 */
-	0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
-	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,	/* 0x30 */
-	0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
-	0x40, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,	/* 0x40 */
-	0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
-	0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77,	/* 0x50 */
-	0x78, 0x79, 0x7a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
-	0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,	/* 0x60 */
-	0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
-	0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77,	/* 0x70 */
-	0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f,
-	0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,	/* 0x80 */
-	0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
-	0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97,	/* 0x90 */
-	0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
-	0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,	/* 0xa0 */
-	0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
-	0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7,	/* 0xb0 */
-	0xb8, 0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf,
-	0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7,	/* 0xc0 */
-	0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
-	0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7,	/* 0xd0 */
-	0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
-	0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7,	/* 0xe0 */
-	0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
-	0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7,	/* 0xf0 */
-	0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff
-};
-static u_char kctypetable[256] = {
-	0200, 0200, 0200, 0200, 0200, 0200, 0200, 0200,	/* 0x00 */
-	0200, 0200, 0200, 0200, 0200, 0200, 0200, 0200,
-	0200, 0200, 0200, 0200, 0200, 0200, 0200, 0200,	/* 0x10 */
-	0200, 0200, 0200, 0200, 0200, 0200, 0200, 0200,
-	   0, 0060, 0060, 0060, 0060, 0060, 0060, 0060,	/* 0x20 */
-	0060, 0060, 0060, 0060, 0060, 0060, 0060, 0060,
-	0060, 0060, 0060, 0060, 0060, 0060, 0060, 0060,	/* 0x30 */
-	0060, 0060, 0060, 0060, 0060, 0060, 0060, 0060,
-	0062, 0062, 0062, 0062, 0062, 0062, 0062, 0062,	/* 0x40 */
-	0062, 0062, 0062, 0062, 0062, 0062, 0062, 0062,
-	0062, 0062, 0062, 0062, 0062, 0062, 0062, 0062,	/* 0x50 */
-	0062, 0062, 0062, 0062, 0062, 0062, 0062, 0062,
-	0022, 0022, 0022, 0022, 0022, 0022, 0022, 0022,	/* 0x60 */
-	0022, 0022, 0022, 0022, 0022, 0022, 0022, 0022,
-	0022, 0022, 0022, 0022, 0022, 0022, 0022, 0022,	/* 0x70 */
-	0022, 0022, 0022, 0022, 0022, 0022, 0022, 0200,
-	0002, 0003, 0003, 0003, 0003, 0003, 0003, 0003,	/* 0x80 */
-	0003, 0003, 0003, 0003, 0003, 0003, 0003, 0003,
-	0003, 0003, 0003, 0003, 0003, 0003, 0003, 0003,	/* 0x90 */
-	0003, 0003, 0003, 0003, 0003, 0003, 0003, 0003,
-	0002, 0016, 0016, 0016, 0016, 0016, 0016, 0016,	/* 0xa0 */
-	0016, 0016, 0016, 0016, 0016, 0016, 0016, 0016,
-	0016, 0016, 0016, 0016, 0016, 0016, 0016, 0016,	/* 0xb0 */
-	0016, 0016, 0016, 0016, 0016, 0016, 0016, 0016,
-	0016, 0016, 0016, 0016, 0016, 0016, 0016, 0016,	/* 0xc0 */
-	0016, 0016, 0016, 0016, 0016, 0016, 0016, 0016,
-	0016, 0016, 0016, 0016, 0016, 0016, 0016, 0016,	/* 0xd0 */
-	0016, 0016, 0016, 0016, 0016, 0016, 0016, 0016,
-	0013, 0013, 0013, 0013, 0013, 0013, 0013, 0013,	/* 0xe0 */
-	0013, 0013, 0013, 0013, 0013, 0013, 0013, 0013,
-	0013, 0013, 0013, 0013, 0013, 0013, 0013, 0013,	/* 0xf0 */
-	0013, 0013, 0013, 0013, 0013, 0010, 0010,    0
-};
-# endif	/* !FD */
-#endif	/* !LSI_C */
+#endif	/* !DJGPP && !_NODOSDRIVE */
 
 
 #ifndef	FD
 static int NEAR _dospath(path)
 char *path;
 {
-	return((isalpha(*path) && path[1] == ':') ? *path : 0);
+	return((isalpha2(*path) && path[1] == ':') ? *path : 0);
 }
 
 static char *NEAR strdelim(s, d)
@@ -305,6 +164,7 @@ int d;
 		if (iskanji1(s, i)) i++;
 # endif
 	}
+
 	return(NULL);
 }
 
@@ -323,6 +183,7 @@ int d;
 		if (iskanji1(s, i)) i++;
 # endif
 	}
+
 	return(cp);
 }
 
@@ -341,6 +202,7 @@ int ptr;
 
 	for (i = 0; s[i] && i < ptr; i++) if (iskanji1(s, i)) i++;
 	if (!s[i] || i > ptr) return(1);
+
 	return(!iskanji1(s, i));
 # else
 	return(1);
@@ -375,6 +237,7 @@ char *s;
 		*cp = _SC_;
 	}
 	*(++cp) = '\0';
+
 	return(cp);
 }
 
@@ -425,27 +288,22 @@ char *buf, *s1, *s2;
 		for (i = 0; s2[i] && i < len; i++) *(cp++) = s2[i];
 	}
 	*cp = '\0';
+
 	return(cp);
 }
 
 int isdotdir(s)
 char *s;
 {
-	if (s[0] == '.' && (!s[1] || (s[1] == '.' && !s[2]))) return(1);
+	if (s[0] != '.') /*EMPTY*/;
+	else if (!s[1]) return(2);
+	else if (s[1] != '.') /*EMPTY*/;
+	else if (!s[2]) return(1);
+
 	return(0);
 }
 #endif	/* !FD */
 
-static int NEAR seterrno(doserr)
-u_short doserr;
-{
-	int i;
-
-	for (i = 0; i < DOSERRLISTSIZ; i++)
-		if (doserr == doserrlist[i]) return(errno = unixerrlist[i]);
-	return(errno = EINVAL);
-}
-
 #ifdef	DJGPP
 static int NEAR dos_putpath(path, offset)
 char *path;
@@ -455,6 +313,7 @@ int offset;
 
 	i = strlen(path) + 1;
 	dosmemput(path, i, __tb + offset);
+
 	return(i);
 }
 #endif
@@ -484,57 +343,10 @@ char *path;
 		}
 	}
 	buf[j] = '\0';
+
 	return(buf);
 }
 
-static long NEAR intcall(vect, regp, sregp)
-int vect;
-__dpmi_regs *regp;
-struct SREGS *sregp;
-{
-#ifdef	__TURBOC__
-	struct REGPACK preg;
-#endif
-
-	(*regp).x.flags |= FR_CARRY;
-#ifdef	__TURBOC__
-	preg.r_ax = (*regp).x.ax;
-	preg.r_bx = (*regp).x.bx;
-	preg.r_cx = (*regp).x.cx;
-	preg.r_dx = (*regp).x.dx;
-	preg.r_bp = (*regp).x.bp;
-	preg.r_si = (*regp).x.si;
-	preg.r_di = (*regp).x.di;
-	preg.r_ds = (*sregp).ds;
-	preg.r_es = (*sregp).es;
-	preg.r_flags = (*regp).x.flags;
-	intr(vect, &preg);
-	(*regp).x.ax = preg.r_ax;
-	(*regp).x.bx = preg.r_bx;
-	(*regp).x.cx = preg.r_cx;
-	(*regp).x.dx = preg.r_dx;
-	(*regp).x.bp = preg.r_bp;
-	(*regp).x.si = preg.r_si;
-	(*regp).x.di = preg.r_di;
-	(*sregp).ds = preg.r_ds;
-	(*sregp).es = preg.r_es;
-	(*regp).x.flags = preg.r_flags;
-#else	/* !__TURBOC__ */
-# ifdef	DJGPP
-	(*regp).x.ds = (*sregp).ds;
-	(*regp).x.es = (*sregp).es;
-	__dpmi_int(vect, regp);
-	(*sregp).ds = (*regp).x.ds;
-	(*sregp).es = (*regp).x.es;
-# else
-	int86x(vect, regp, regp, sregp);
-# endif
-#endif	/* !__TURBOC__ */
-	if (!((*regp).x.flags & FR_CARRY)) return((long)((*regp).x.ax));
-	seterrno((*regp).x.ax);
-	return(-1L);
-}
-
 int getcurdrv(VOID_A)
 {
 	int drive;
@@ -545,6 +357,7 @@ int getcurdrv(VOID_A)
 	else
 #endif
 	drive += (bdos(0x19, 0, 0) & 0xff);
+
 	return(drive);
 }
 
@@ -557,7 +370,7 @@ int drive, nodir;
 	int drv, olddrv;
 
 	errno = EINVAL;
-	if (drive >= 'a' && drive <= 'z') {
+	if (islower2(drive)) {
 		drv = drive - 'a';
 		dos7access &= ~D7_CAPITAL;
 	}
@@ -587,7 +400,7 @@ int drive, nodir;
 	}
 
 	if (!nodir) {
-		path = ".";
+		path = curpath;
 		reg.x.ax = 0x3b00;
 #ifdef	DJGPP
 		dos_putpath(path, 0);
@@ -601,10 +414,10 @@ int drive, nodir;
 			return(-1);
 		}
 	}
-
 #ifndef	_NODOSDRIVE
 	lastdrive = drive + 'a';
 #endif
+
 	return(0);
 }
 
@@ -629,6 +442,7 @@ int getdosver(VOID_A)
 		}
 		dos7access |= (ver & D7_DOSVER);
 	}
+
 	return(dos7access & D7_DOSVER);
 }
 
@@ -655,12 +469,11 @@ char *path;
 #endif
 	if (getdosver() < 7) {
 #ifndef	_NODOSDRIVE
-		if ((dosdrive & 1) && drv[0] >= 'a' && drv[0] <= 'z')
-			return(-1);
+		if ((dosdrive & 1) && islower2(drv[0])) return(-1);
 #endif
 		return(-3);
 	}
-	if (drv[0] >= 'A' && drv[0] <= 'Z') return(0);
+	if (isupper2(drv[0])) return(0);
 
 	drv[1] = ':';
 	drv[2] = '\\';
@@ -687,6 +500,7 @@ char *path;
 	dosmemget(__tb + sizeof(drv), sizeof(buf), buf);
 #endif
 	if (!strcmp(buf, VOL_FAT32)) return(2);
+
 	return(1);
 }
 #endif	/* !_NOUSELFN */
@@ -713,6 +527,7 @@ int drive;
 #ifndef	BSPATHDELIM
 	adjustpname(pathname);
 #endif
+
 	return(pathname);
 }
 
@@ -735,6 +550,7 @@ char *path, *buf;
 	}
 	else if (cp == &(path[2])) return(path);
 	else strcpy(&(buf[2]), cp);
+
 	return(buf);
 }
 # endif	/* !_NODOSDRIVE */
@@ -778,7 +594,7 @@ char *path, *alias;
 # endif
 
 	if (!path || !(i = _dospath(path))) i = getcurdrv();
-	if (i >= 'a' && i <= 'z') *alias = tolower2(*alias);
+	if (islower2(i)) *alias = tolower2(*alias);
 
 	return(alias);
 }
@@ -849,8 +665,9 @@ char *path, *resolved;
 			resolved[0] = i;
 			resolved[1] = ':';
 			i = 2;
-			if (isalpha(resolved[2])
-			&& resolved[3] == '.' && resolved[4] == _SC_) i = 5;
+			if (isalpha2(resolved[2])
+			&& resolved[3] == '.' && resolved[4] == _SC_)
+				i = 5;
 			while (resolved[i]) if (resolved[i++] == _SC_) break;
 			resolved[2] = _SC_;
 			for (j = 0; resolved[i + j]; j++)
@@ -866,7 +683,7 @@ char *path, *resolved;
 				strcpy(&(resolved[2]), path);
 		}
 	}
-	else *resolved = (i >= 'A' && i <= 'Z')
+	else *resolved = (isupper2(i))
 		? toupper2(*resolved) : tolower2(*resolved);
 
 	return(resolved);
@@ -882,6 +699,7 @@ char *path;
 		if (path[i] == '\\') path[i] = _SC_;
 		else if (iskanji1(path, i)) i++;
 	}
+
 	return(path);
 }
 #endif	/* !BSPATHDELIM */
@@ -908,6 +726,7 @@ char *path;
 # endif
 		else path[i] = toupper2(path[i]);
 	}
+
 	return(path);
 }
 #endif	/* DJGPP || !BSPATHDELIM */
@@ -930,6 +749,7 @@ char *path, *alias;
 #ifndef	_NODOSDRIVE
 	else if (i < 0 && errno == EACCES) return(path);
 #endif
+
 	return(NULL);
 }
 
@@ -1060,6 +880,7 @@ int nsect, mode, retry;
 	}
 	if (cp != buf) free(cp);
 	errno = EACCES;
+
 	return(-1);
 }
 
@@ -1176,6 +997,7 @@ int nsect, mode, retry;
 	}
 	if (cp != buf) free(cp);
 	errno = EACCES;
+
 	return(-1);
 }
 #endif	/* !PC98 */
@@ -1264,7 +1086,7 @@ u_long l_sect, e_sect;
 		if (PART_TABLE + ofs >= size) {
 #ifndef	PC98
 			if (ps) {
-				if (xbiosdiskio(maxdrive, ps++, 0L,
+				if (xbiosdiskio(maxdrive, ps++, (u_long)0,
 				buf, 1, BIOS_XREAD, BIOSRETRY) < 0
 				&& ((drvlist[maxdrive].flags & DI_INVALIDCHS)
 				|| _biosdiskio(maxdrive, head, sect, cyl,
@@ -1314,14 +1136,15 @@ u_long l_sect, e_sect;
 			continue;
 		}
 		if (pt -> filesys == PT_EXTEND) {
-			if (_checkdrive(sh, ss, sc, 0L, 0L) < 0) {
+			if (_checkdrive(sh, ss, sc, (u_long)0, (u_long)0) < 0)
+			{
 				free(buf);
 				return(-1);
 			}
 			continue;
 		}
 
-		drvlist[maxdrive].f_sect = 0L;
+		drvlist[maxdrive].f_sect = (u_long)0;
 		if (pt -> filesys == PT_FAT16XLBA
 		|| pt -> filesys == PT_FAT32LBA)
 			drvlist[maxdrive].f_sect = ls + l_sect;
@@ -1347,6 +1170,7 @@ u_long l_sect, e_sect;
 		drvlist[maxdrive].flags = drvlist[maxdrive - 1].flags;
 	}
 	free(buf);
+
 	return(1);
 }
 
@@ -1383,14 +1207,18 @@ int drive;
 
 #ifdef	PC98
 		for (i = 0; i < MAX_HDD && maxdrive < 'Z' - 'A' + 1; i++) {
-			if (getdrvparam(BIOS_HDD | i,
-				&(drvlist[maxdrive])) <= 0) continue;
-			if (_checkdrive(0, 1, 0, 0L, 0L) < 0) return(-1);
+			if (getdrvparam(BIOS_HDD | i, &(drvlist[maxdrive]))
+			<= 0)
+				continue;
+			if (_checkdrive(0, 1, 0, (u_long)0, (u_long)0) < 0)
+				return(-1);
 		}
 		for (i = 0; i < MAX_SCSI && maxdrive < 'Z' - 'A' + 1; i++) {
-			if (getdrvparam(BIOS_SCSI | i,
-				&(drvlist[maxdrive])) <= 0) continue;
-			if (_checkdrive(0, 1, 0, 0L, 0L) < 0) return(-1);
+			if (getdrvparam(BIOS_SCSI | i, &(drvlist[maxdrive]))
+			<= 0)
+				continue;
+			if (_checkdrive(0, 1, 0, (u_long)0, (u_long)0) < 0)
+				return(-1);
 		}
 #else
 		reg.h.ah = BIOS_PARAM;
@@ -1399,9 +1227,10 @@ int drive;
 		n = reg.h.dl;
 
 		for (i = 0; i < n && maxdrive < 'Z' - 'A' + 1; i++) {
-			if (!(j = getdrvparam(BIOS_HDD | i,
-				&(drvlist[maxdrive])))) continue;
-			if (j < 0 || _checkdrive(0, 0, 0, 0L, 0L) < 0)
+			j = getdrvparam(BIOS_HDD | i, &(drvlist[maxdrive]));
+			if (!j) continue;
+			if (j < 0
+			|| _checkdrive(0, 0, 0, (u_long)0, (u_long)0) < 0)
 				return(-1);
 		}
 #endif
@@ -1409,6 +1238,7 @@ int drive;
 	if (drive >= 0 && drive < maxdrive
 	&& (drvlist[drive].flags & DI_PSEUDO))
 		return(drvlist[drive].filesys);
+
 	return(0);
 }
 
@@ -1454,10 +1284,12 @@ int n, size, iswrite;
 	head = sect / drvlist[drive].sect;
 	sect %= drvlist[drive].sect;
 
-	if (!iswrite) return(_biosdiskio(drive, head, sect, cyl,
-		buf, n, BIOS_READ, BIOSRETRY));
+	if (!iswrite)
+		return(_biosdiskio(drive, head, sect, cyl,
+			buf, n, BIOS_READ, BIOSRETRY));
 	else if (_biosdiskio(drive, head, sect, cyl,
-	buf, n, BIOS_WRITE, BIOSRETRY) < 0) return(-1);
+	buf, n, BIOS_WRITE, BIOSRETRY) < 0)
+		return(-1);
 	else return(_biosdiskio(drive, head, sect, cyl,
 		buf, n, BIOS_VERIFY, 1));
 }
@@ -1496,6 +1328,7 @@ int drv;
 			return(-1);
 		}
 	}
+
 	return(end);
 }
 
@@ -1517,6 +1350,7 @@ int drv, level;
 		reg.x.cx = 0x086a;
 		if (int21call(&reg, &sreg) < 0) return(-1);
 	}
+
 	return(0);
 }
 
@@ -1639,6 +1473,7 @@ int n, size, iswrite;
 #ifdef	DJGPP
 	if (!iswrite) dosmemget(__tb + sizeof(struct iopacket_t), size, buf);
 #endif
+
 	return(0);
 }
 
@@ -1690,6 +1525,7 @@ int n, size, iswrite;
 		ret = dosrawdiskio(drv, sect, buf, n, size, iswrite);
 		if (iswrite && unlockdrive(drv, level) < 0) return(-1);
 	}
+
 	return(ret);
 }
 #endif	/* !_NODOSDRIVE */
@@ -1697,7 +1533,7 @@ int n, size, iswrite;
 
 static int NEAR dos_findfirst(path, attr, result)
 char *path;
-u_short attr;
+u_int attr;
 struct dosfind_t *result;
 {
 	struct SREGS sreg;
@@ -1720,6 +1556,7 @@ struct dosfind_t *result;
 #ifdef	DJGPP
 	dosmemget(__tb, sizeof(struct dosfind_t), result);
 #endif
+
 	return(0);
 }
 
@@ -1742,18 +1579,19 @@ struct dosfind_t *result;
 #ifdef	DJGPP
 	dosmemget(__tb, sizeof(struct dosfind_t), result);
 #endif
+
 	return(0);
 }
 
 #ifndef	_NOUSELFN
-static long NEAR lfn_findfirst(path, attr, result)
+static int NEAR lfn_findfirst(fdp, path, attr, result)
+u_int *fdp;
 char *path;
-u_short attr;
+u_int attr;
 struct lfnfind_t *result;
 {
 	struct SREGS sreg;
 	__dpmi_regs reg;
-	long fd;
 
 	path = duplpath(path);
 	reg.x.ax = 0x714e;
@@ -1766,15 +1604,17 @@ struct lfnfind_t *result;
 	reg.x.dx = PTR_OFF(path, sizeof(struct lfnfind_t));
 	sreg.es = PTR_SEG(result);
 	reg.x.di = PTR_OFF(result, 0);
-	if ((fd = int21call(&reg, &sreg)) < 0) return(-1L);
+	if (int21call(&reg, &sreg) < 0) return(-1);
+	*fdp = reg.x.ax;
 #ifdef	DJGPP
 	dosmemget(__tb, sizeof(struct lfnfind_t), result);
 #endif
-	return(fd);
+
+	return(0);
 }
 
 static int NEAR lfn_findnext(fd, result)
-u_short fd;
+u_int fd;
 struct lfnfind_t *result;
 {
 	struct SREGS sreg;
@@ -1789,18 +1629,20 @@ struct lfnfind_t *result;
 #ifdef	DJGPP
 	dosmemget(__tb, sizeof(struct lfnfind_t), result);
 #endif
+
 	return(0);
 }
 
 static int NEAR lfn_findclose(fd)
-u_short fd;
+u_int fd;
 {
 	struct SREGS sreg;
 	__dpmi_regs reg;
 
 	reg.x.ax = 0x71a1;
 	reg.x.bx = fd;
-	return((int21call(&reg, &sreg) < 0) ? -1 : 0);
+
+	return(int21call(&reg, &sreg));
 }
 #endif	/* !_NOUSELFN */
 
@@ -1839,18 +1681,12 @@ char *dir;
 			(struct dosfind_t *)(dirp -> dd_buf));
 		if (i >= 0) dirp -> dd_id |= DID_IFLABEL;
 		else {
-			long fd;
+			u_int fd;
 
-			fd = lfn_findfirst(path, SEARCHATTRS,
+			i = lfn_findfirst(&fd, path, SEARCHATTRS,
 				(struct lfnfind_t *)(dirp -> dd_buf));
-			if (fd >= 0) {
-				dirp -> dd_fd = (u_short)fd;
-				i = 0;
-			}
-			else {
-				dirp -> dd_id &= ~DID_IFLFN;
-				i = -1;
-			}
+			if (i >= 0) dirp -> dd_fd = fd;
+			else dirp -> dd_id &= ~DID_IFLFN;
 		}
 	}
 	else
@@ -1892,6 +1728,7 @@ char *dir;
 		return(NULL);
 	}
 	strcpy(dirp -> dd_path, dir);
+
 	return(dirp);
 }
 
@@ -1907,6 +1744,7 @@ DIR *dirp;
 	free(dirp -> dd_buf);
 	free(dirp -> dd_path);
 	free(dirp);
+
 	return(0);
 }
 
@@ -1955,19 +1793,13 @@ DIR *dirp;
 
 #ifndef	_NOUSELFN
 		if (dirp -> dd_id & DID_IFLFN) {
-			long fd;
+			u_int fd;
 
 			strcpy(cp, "*");
-			fd = lfn_findfirst(path, SEARCHATTRS,
+			i = lfn_findfirst(&fd, path, SEARCHATTRS,
 				(struct lfnfind_t *)(dirp -> dd_buf));
-			if (fd >= 0) {
-				dirp -> dd_fd = (u_short)fd;
-				i = 0;
-			}
-			else {
-				dirp -> dd_id &= ~DID_IFLFN;
-				i = -1;
-			}
+			if (i >= 0) dirp -> dd_fd = fd;
+			else dirp -> dd_id &= ~DID_IFLFN;
 		}
 		else
 #endif	/* !_NOUSELFN */
@@ -2026,6 +1858,7 @@ int mode;
 		errno = EEXIST;
 		return(-1);
 	}
+
 	return(mkdir(path) ? -1 : 0);
 }
 # endif
@@ -2057,7 +1890,8 @@ char *path;
 #endif
 	sreg.ds = PTR_SEG(path);
 	reg.x.dx = PTR_OFF(path, 0);
-	return((int21call(&reg, &sreg) < 0) ? -1 : 0);
+
+	return(int21call(&reg, &sreg));
 }
 
 static int NEAR gendosname(s)
@@ -2087,6 +1921,7 @@ char *s;
 	}
 	if (i) *(s++) = '.';
 	s[i] = '\0';
+
 	return(1);
 }
 
@@ -2101,7 +1936,7 @@ char *from, *to;
 	strcpy(fbuf, to);
 	if (!(tp = strrdelim(fbuf, 1))) {
 		tp = to;
-		strcpy(fbuf, ".");
+		strcpy(fbuf, curpath);
 	}
 	else {
 		if (*tp == _SC_) tp++;
@@ -2109,7 +1944,8 @@ char *from, *to;
 		tp = &(to[tp - fbuf]);
 	}
 	if (!(to = unixrealpath(fbuf, tbuf))
-	|| !(from = unixrealpath(from, fbuf))) return(-1);
+	|| !(from = unixrealpath(from, fbuf)))
+		return(-1);
 	strcpy(strcatdelim(tbuf), tp);
 
 	for (i = 0; from[i]; i++) if (from[i] != to[i]) break;
@@ -2184,7 +2020,8 @@ char *from, *to;
 		if (i == 1) {
 			if (ax != 0x5600
 			|| dos_findfirst(from, SEARCHATTRS, &dbuf) < 0
-			|| !(dbuf.attr & DS_IFDIR)) break;
+			|| !(dbuf.attr & DS_IFDIR))
+				break;
 
 			if (t > -3) {
 				if ((f = gendosname(to)) < 0) break;
@@ -2198,6 +2035,7 @@ char *from, *to;
 		}
 	}
 	errno = EACCES;
+
 	return(-1);
 }
 
@@ -2229,6 +2067,7 @@ int mode;
 	reg.x.dx = PTR_OFF(path, 0);
 	if (int21call(&reg, &sreg) >= 0) return(0);
 	if (reg.x.ax == 5) errno = EEXIST;
+
 	return(-1);
 }
 
@@ -2254,7 +2093,8 @@ char *path;
 #endif
 	sreg.ds = PTR_SEG(path);
 	reg.x.dx = PTR_OFF(path, 0);
-	return((int21call(&reg, &sreg) < 0) ? -1 : 0);
+
+	return(int21call(&reg, &sreg));
 }
 
 int unixchdir(path)
@@ -2280,7 +2120,8 @@ char *path;
 #endif
 	sreg.ds = PTR_SEG(path);
 	reg.x.dx = PTR_OFF(path, 0);
-	return((int21call(&reg, &sreg) < 0) ? -1 : 0);
+
+	return(int21call(&reg, &sreg));
 }
 #endif	/* !_NOUSELFN */
 
@@ -2314,6 +2155,7 @@ int size;
 #ifdef	DJGPP
 	if (unixrealpath(pathname, tmp)) strcpy(pathname, tmp);
 #endif
+
 	return(pathname);
 }
 
@@ -2409,14 +2251,13 @@ statfs_t *buf;
 	i = supportLFN(path);
 # ifndef	_NODOSDRIVE
 	if (i == -2) {
-		char tmp[sizeof(long) * 3 + 1];
+		char tmp[3 * sizeof(long) + 1];
 
 		if (dosstatfs(drive, tmp) < 0) return(-1);
-		buf -> f_bsize = *((u_long *)&(tmp[sizeof(long) * 0]));
-		buf -> f_blocks = *((u_long *)&(tmp[sizeof(long) * 1]));
+		buf -> f_bsize = *((u_long *)&(tmp[0 * sizeof(long)]));
+		buf -> f_blocks = *((u_long *)&(tmp[1 * sizeof(long)]));
 		buf -> f_bfree =
-		buf -> f_bavail = *((u_long *)&(tmp[sizeof(long) * 2]));
-		buf -> f_files = -1;
+		buf -> f_bavail = *((u_long *)&(tmp[2 * sizeof(long)]));
 	}
 	else
 # endif
@@ -2431,6 +2272,7 @@ statfs_t *buf;
 		sreg.ds = PTR_SEG(path);
 		reg.x.dx = PTR_OFF(path, sizeof(fsbuf));
 		if (int21call(&reg, &sreg) < 0) return(-1);
+
 # ifdef	DJGPP
 		dosmemget(__tb, sizeof(fsbuf), &fsbuf);
 # endif
@@ -2439,7 +2281,6 @@ statfs_t *buf;
 		buf -> f_blocks = (u_long)fsbuf.f_blocks;
 		buf -> f_bfree =
 		buf -> f_bavail = (u_long)fsbuf.f_bavail;
-		buf -> f_files = -1;
 	}
 	else
 #endif	/* !_NOUSELFN */
@@ -2452,19 +2293,19 @@ statfs_t *buf;
 			return(-1);
 		}
 
-		if (!reg.x.ax || !reg.x.cx || !reg.x.dx || reg.x.dx < reg.x.bx)
-			buf -> f_bsize = buf -> f_blocks =
-			buf -> f_bfree = buf -> f_bavail = -1L;
-		else {
-			buf -> f_bsize = (u_long)(reg.x.ax)
-				* (u_long)(reg.x.cx);
-			buf -> f_blocks = (u_long)(reg.x.dx);
-			buf -> f_bfree =
-			buf -> f_bavail = (u_long)(reg.x.bx);
-		}
-		buf -> f_files = -1;
+		buf -> f_bsize = (u_long)(reg.x.ax) * (u_long)(reg.x.cx);
+		buf -> f_blocks = (u_long)(reg.x.dx);
+		buf -> f_bfree =
+		buf -> f_bavail = (u_long)(reg.x.bx);
 	}
 
+	if (!(buf -> f_blocks) || !(buf -> f_bsize)
+	|| buf -> f_blocks < buf -> f_bavail) {
+		errno = EIO;
+		return(-1);
+	}
+	buf -> f_files = -1L;
+
 	return(0);
 }
 
@@ -2474,7 +2315,7 @@ struct stat *stp;
 {
 #ifndef	_NOUSELFN
 	struct lfnfind_t lbuf;
-	long fd;
+	u_int fd;
 #endif
 	struct dosfind_t dbuf;
 	int i;
@@ -2482,8 +2323,8 @@ struct stat *stp;
 #ifdef	_NOUSELFN
 	if ((i = dos_findfirst(path, SEARCHATTRS, &dbuf)) < 0) {
 		if (errno && errno != ENOENT) return(-1);
-		if (!strcmp(path, ".."))
-			i = dos_findfirst(".", SEARCHATTRS, &dbuf);
+		if (isdotdir(path) == 1)
+			i = dos_findfirst(curpath, SEARCHATTRS, &dbuf);
 		else i = dos_findfirst(path, DS_IFLABEL, &dbuf);
 	}
 #else	/* !_NOUSELFN */
@@ -2498,25 +2339,21 @@ struct stat *stp;
 	}
 # endif	/* !_NODOSDRIVE */
 	if (i <= 0) {
-		fd = -1;
+		fd = (u_int)-1;
 		if ((i = dos_findfirst(path, SEARCHATTRS, &dbuf)) < 0) {
 			if (errno && errno != ENOENT) return(-1);
-			if (!strcmp(path, ".."))
-				i = dos_findfirst(".", SEARCHATTRS, &dbuf);
+			if (isdotdir(path) == 1)
+				i = dos_findfirst(curpath, SEARCHATTRS, &dbuf);
 			else i = dos_findfirst(path, DS_IFLABEL, &dbuf);
 		}
 	}
 	else {
-		if ((fd = lfn_findfirst(path, SEARCHATTRS, &lbuf)) < 0) {
+		if ((i = lfn_findfirst(&fd, path, SEARCHATTRS, &lbuf)) < 0) {
 			if (errno && errno != ENOENT) return(-1);
-			if (!strcmp(path, "..")) {
-				fd = lfn_findfirst(".", SEARCHATTRS, &lbuf);
-				i = (fd >= 0) ? 0 : -1;
-			}
-			else {
-				fd = -1;
-				i = dos_findfirst(path, DS_IFLABEL, &dbuf);
-			}
+			if (isdotdir(path) == 1)
+				i = lfn_findfirst(&fd, curpath,
+					SEARCHATTRS, &lbuf);
+			else i = dos_findfirst(path, DS_IFLABEL, &dbuf);
 		}
 	}
 #endif	/* !_NOUSELFN */
@@ -2526,7 +2363,7 @@ struct stat *stp;
 	}
 
 #ifndef	_NOUSELFN
-	if (fd >= 0) {
+	if (fd != (u_int)-1) {
 		stp -> st_mode = getdosmode(lbuf.attr);
 		stp -> st_mtime = getdostime(lbuf.wrdate, lbuf.wrtime);
 		stp -> st_size = lbuf.size_l;
@@ -2543,6 +2380,7 @@ struct stat *stp;
 	stp -> st_dev = stp -> st_ino = 0;
 	stp -> st_nlink = 1;
 	stp -> st_uid = stp -> st_gid = -1;
+
 	return(0);
 }
 
@@ -2580,7 +2418,8 @@ int mode;
 #endif
 	sreg.ds = PTR_SEG(path);
 	reg.x.dx = PTR_OFF(path, 0);
-	return((int21call(&reg, &sreg) < 0) ? -1 : 0);
+
+	return(int21call(&reg, &sreg));
 }
 
 #ifndef	_NOUSELFN
@@ -2589,21 +2428,20 @@ int unixutime(path, times)
 char *path;
 struct utimbuf *times;
 {
-	time_t t = times -> modtime;
+	time_t t;
 	__dpmi_regs reg;
 	struct SREGS sreg;
 	char buf[MAXPATHLEN];
-	int i;
+	int i, fd;
 
+	t = times -> modtime;
 	path = duplpath(path);
 	i = supportLFN(path);
-	if (i <= -3) return(utime(path, times));
-	else if (i < 0) {
+	if (i < 0 && i > -3) {
 # ifndef	_NODOSDRIVE
 		return(dosutime(regpath(path, buf), times));
 # else
 		if (!(path = preparefile(path, buf))) return(-1);
-		return(utime(path, times));
 # endif
 	}
 #else	/* !USEUTIME */
@@ -2611,24 +2449,32 @@ int unixutimes(path, tvp)
 char *path;
 struct timeval tvp[2];
 {
-	time_t t = tvp[1].tv_sec;
+	time_t t;
 	__dpmi_regs reg;
 	struct SREGS sreg;
 	char buf[MAXPATHLEN];
-	int i;
+	int i, fd;
 
+	t = tvp[1].tv_sec;
 	path = duplpath(path);
 	i = supportLFN(path);
-	if (i <= -3) return(utimes(path, tvp));
-	else if (i < 0) {
+	if (i < 0 && i > -3) {
 # ifndef	_NODOSDRIVE
 		return(dosutimes(regpath(path, buf), tvp));
 # else
 		if (!(path = preparefile(path, buf))) return(-1);
-		return(utimes(path, tvp));
 # endif
 	}
 #endif	/* !USEUTIME */
+	if (i <= 0) {
+		if ((fd = open(path, O_RDONLY, 0666)) < 0) return(-1);
+		reg.x.ax = 0x5701;
+		reg.x.bx = (u_short)fd;
+		putdostime(&(reg.x.dx), &(reg.x.cx), t);
+		i = int21call(&reg, &sreg);
+		close(fd);
+		return(i);
+	}
 
 	reg.x.ax = 0x7143;
 	reg.h.bl = 0x03;
@@ -2638,7 +2484,8 @@ struct timeval tvp[2];
 #endif
 	sreg.ds = PTR_SEG(path);
 	reg.x.dx = PTR_OFF(path, 0);
-	return((int21call(&reg, &sreg) < 0) ? -1 : 0);
+
+	return(int21call(&reg, &sreg));
 }
 
 int unixopen(path, flags, mode)
@@ -2689,6 +2536,7 @@ int flags, mode;
 	int21call(&reg, &sreg);
 
 	if (!(path = shortname(path, buf))) return(-1);
+
 	return(open(path, flags, mode));
 }
 
@@ -2736,6 +2584,7 @@ char *path, *type;
 	}
 
 	if ((i = unixopen(path, flags, 0666)) < 0) return(NULL);
+
 	return(fdopen(i, type));
 }
 #endif	/* !_NOUSELFN */
